# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import brainstate
import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import scipy.sparse as sp

import brainevent
from brainevent._csr.plasticity_binary import (
    update_csr_on_binary_pre,
    update_csr_on_binary_post,
    update_csr_on_binary_pre_p,
    update_csr_on_binary_post_p,
)

PLATFORM = jax.default_backend()
PRE_BACKENDS = tuple(update_csr_on_binary_pre_p.available_backends(PLATFORM))
POST_BACKENDS = tuple(update_csr_on_binary_post_p.available_backends(PLATFORM))

shapes = [
    (20, 100),
    (100, 100),
    (100, 50),
]


class Test_csr_on_pre:
    @pytest.mark.parametrize('backend', PRE_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    def test_csr_on_pre_v1(self, backend, shape):
        n_pre, n_post = shape
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)

        pre_spike = brainstate.random.random((n_pre,)) < 0.5
        post_trace = brainstate.random.random((n_post,))

        csr = brainevent.CSR.fromdense(mat)
        csr2 = csr.with_data(
            update_csr_on_binary_pre(csr.data, csr.indices, csr.indptr, pre_spike, post_trace,
                                     shape=csr.shape, backend=backend)
        )
        dense2 = jnp.where(mask, mat + jnp.outer(pre_spike.astype(float), post_trace), 0.)

        assert jnp.allclose(csr2.todense(), dense2)

        jax.block_until_ready((mat, pre_spike, post_trace))

    @pytest.mark.parametrize('backend', PRE_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('mat_unit', [u.mV, u.ms])
    @pytest.mark.parametrize('trace_unit', [u.mV, u.ms])
    def test_csr_on_pre_with_unit(self, backend, shape, mat_unit, trace_unit):
        def run():
            n_pre, n_post = shape
            mat = brainstate.random.random((n_pre, n_post))
            mask = mat < 0.5
            mat = jnp.where(mask, mat, 0.) * mat_unit
            pre_spike = brainstate.random.random((n_pre,)) < 0.1
            post_trace = brainstate.random.random((n_post,)) * trace_unit

            csr = brainevent.CSR.fromdense(mat)
            csr = csr.with_data(
                update_csr_on_binary_pre(csr.data, csr.indices, csr.indptr, pre_spike, post_trace,
                                         shape=csr.shape, backend=backend)
            )

            dense = mat + u.math.outer(pre_spike.astype(float), post_trace)
            dense = u.math.where(mask, dense, 0. * mat_unit)

            assert u.math.allclose(csr.todense(), dense)

            jax.block_until_ready((mat, pre_spike, post_trace))

        if mat_unit.has_same_dim(trace_unit):
            run()
        else:
            with pytest.raises(u.UnitMismatchError):
                run()

    @pytest.mark.parametrize('backend', PRE_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('w_in', [None, 0.1])
    @pytest.mark.parametrize('w_max', [None, 0.5])
    def test_csr_on_pre_v2(self, backend, shape, w_in, w_max):
        n_pre, n_post = shape
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        pre_spike = brainstate.random.random((n_pre,)) < 0.1
        post_trace = brainstate.random.random((n_post,))

        csr = brainevent.CSR.fromdense(mat)
        csr = csr.with_data(
            update_csr_on_binary_pre(csr.data, csr.indices, csr.indptr, pre_spike, post_trace,
                                     w_min=w_in, w_max=w_max, shape=csr.shape, backend=backend)
        )

        mat = mat + jnp.outer(pre_spike.astype(float), post_trace)
        mat = u.math.clip(mat, a_min=w_in, a_max=w_max)

        mat = jnp.where(mask, mat, 0.)
        assert jnp.allclose(csr.todense(), mat)

        jax.block_until_ready((mat, pre_spike, post_trace))


def _csr_to_csc_with_weight_indices(csr_data, csr_indices, csr_indptr, shape):
    """Convert CSR format to CSC format and return weight indices mapping.

    Returns:
        csc_indices: Row indices in CSC format (presynaptic neuron indices)
        csc_indptr: Column pointers in CSC format
        weight_indices: Mapping from CSC positions to original CSR data positions
    """
    # Handle Quantity data by extracting mantissa
    if hasattr(csr_data, 'mantissa'):
        csr_data_np = np.asarray(csr_data.mantissa)
    else:
        csr_data_np = np.asarray(csr_data)

    # Create scipy CSR matrix
    scipy_csr = sp.csr_matrix(
        (csr_data_np, np.asarray(csr_indices), np.asarray(csr_indptr)),
        shape=shape
    )
    # Convert to CSC
    scipy_csc = scipy_csr.tocsc()

    # The weight_indices maps CSC position -> CSR position
    # We can compute this by tracking where each element came from
    # CSC data comes from reordering CSR data based on column sorting

    # Create a CSR matrix with data = position indices
    position_indices = np.arange(len(csr_data_np), dtype=np.int32)
    scipy_csr_indices = sp.csr_matrix(
        (position_indices, np.asarray(csr_indices), np.asarray(csr_indptr)),
        shape=shape
    )
    scipy_csc_indices = scipy_csr_indices.tocsc()
    weight_indices = scipy_csc_indices.data

    return (
        jnp.asarray(scipy_csc.indices, dtype=jnp.int32),
        jnp.asarray(scipy_csc.indptr, dtype=jnp.int32),
        jnp.asarray(weight_indices, dtype=jnp.int32)
    )


class Test_on_post:
    @pytest.mark.parametrize('backend', POST_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    def test_csr_on_post_v1(self, backend, shape):
        n_pre, n_post = shape
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)

        post_spike = brainstate.random.random((n_post,)) < 0.5
        pre_trace = brainstate.random.random((n_pre,))

        csr = brainevent.CSR.fromdense(mat)
        # Convert CSR to CSC format for csr2csc_on_post
        csc_indices, csc_indptr, weight_indices = _csr_to_csc_with_weight_indices(
            csr.data, csr.indices, csr.indptr, csr.shape
        )

        new_weights = update_csr_on_binary_post(
            csr.data, csc_indices, csc_indptr, weight_indices,
            pre_trace, post_spike, shape=csr.shape, backend=backend
        )
        csr2 = csr.with_data(new_weights)
        dense2 = jnp.where(mask, mat + jnp.outer(pre_trace, post_spike.astype(float)), 0.)

        assert jnp.allclose(csr2.todense(), dense2)

        jax.block_until_ready((mat, post_spike, pre_trace))

    @pytest.mark.parametrize('backend', POST_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('mat_unit', [u.mV, u.ms])
    @pytest.mark.parametrize('trace_unit', [u.mV, u.ms])
    def test_csr_on_post_with_unit(self, backend, shape, mat_unit, trace_unit):
        def run():
            n_pre, n_post = shape
            mat = brainstate.random.random((n_pre, n_post))
            mask = mat < 0.5
            mat = jnp.where(mask, mat, 0.) * mat_unit
            post_spike = brainstate.random.random((n_post,)) < 0.1
            pre_trace = brainstate.random.random((n_pre,)) * trace_unit

            csr = brainevent.CSR.fromdense(mat)
            # Convert CSR to CSC format for csr2csc_on_post
            csc_indices, csc_indptr, weight_indices = _csr_to_csc_with_weight_indices(
                csr.data, csr.indices, csr.indptr, csr.shape
            )

            new_weights = update_csr_on_binary_post(
                csr.data, csc_indices, csc_indptr, weight_indices,
                pre_trace, post_spike, shape=csr.shape, backend=backend
            )
            csr = csr.with_data(new_weights)

            dense = mat + u.math.outer(pre_trace, post_spike.astype(float))
            dense = u.math.where(mask, dense, 0. * mat_unit)

            assert u.math.allclose(csr.todense(), dense)

            jax.block_until_ready((mat, post_spike, pre_trace))

        if mat_unit.has_same_dim(trace_unit):
            run()
        else:
            with pytest.raises(u.UnitMismatchError):
                run()

    @pytest.mark.parametrize('backend', POST_BACKENDS)
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('w_in', [None, 0.1])
    @pytest.mark.parametrize('w_max', [None, 0.5])
    def test_csr_on_post_v2(self, backend, shape, w_in, w_max):
        n_pre, n_post = shape
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        post_spike = brainstate.random.random((n_post,)) < 0.1
        pre_trace = brainstate.random.random((n_pre,))

        csr = brainevent.CSR.fromdense(mat)
        # Convert CSR to CSC format for csr2csc_on_post
        csc_indices, csc_indptr, weight_indices = _csr_to_csc_with_weight_indices(
            csr.data, csr.indices, csr.indptr, csr.shape
        )

        new_weights = update_csr_on_binary_post(
            csr.data, csc_indices, csc_indptr, weight_indices,
            pre_trace, post_spike, w_min=w_in, w_max=w_max, shape=csr.shape, backend=backend
        )
        csr = csr.with_data(new_weights)

        mat = mat + jnp.outer(pre_trace, post_spike.astype(float))
        mat = u.math.clip(mat, a_min=w_in, a_max=w_max)

        mat = jnp.where(mask, mat, 0.)
        assert jnp.allclose(csr.todense(), mat)

        jax.block_until_ready((mat, post_spike, pre_trace))



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import brainstate
import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import brainevent
from brainevent._coo.plasticity_binary import (
    update_coo_on_binary_pre,
    update_coo_on_binary_post,
    update_coo_on_binary_pre_p,
    update_coo_on_binary_post_p,
)

PLATFORM = jax.default_backend()
PRE_BACKENDS = tuple(update_coo_on_binary_pre_p.available_backends(PLATFORM))
POST_BACKENDS = tuple(update_coo_on_binary_post_p.available_backends(PLATFORM))


class Test_coo_on_pre:
    def test_coo_on_pre_v1(self):
        n_pre = 100
        n_post = 100
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        pre_spike = brainstate.random.random((n_pre,)) < 0.1
        post_trace = brainstate.random.random((n_post,))

        coo = brainevent.COO.fromdense(mat)
        coo = coo.with_data(update_coo_on_binary_pre(coo.data, coo.row, coo.col, pre_spike, post_trace))

        mat = mat + jnp.outer(pre_spike, post_trace)
        mat = jnp.where(mask, mat, 0.)

        assert jnp.allclose(coo.todense(), mat, rtol=1e-2, atol=1e-2)
        jax.block_until_ready((mat, pre_spike, post_trace))

    @pytest.mark.parametrize('mat_unit', [u.mV, u.ms])
    @pytest.mark.parametrize('trace_unit', [u.mV, u.ms])
    def test_coo_on_pre_unit(self, mat_unit, trace_unit):
        def run():
            n_pre = 100
            n_post = 100
            mat = brainstate.random.random((n_pre, n_post))
            mask = mat < 0.5
            mat = jnp.where(mask, mat, 0.) * mat_unit
            pre_spike = brainstate.random.random((n_pre,)) < 0.1
            post_trace = brainstate.random.random((n_post,)) * trace_unit

            coo = brainevent.COO.fromdense(mat)
            coo = coo.with_data(update_coo_on_binary_pre(coo.data, coo.row, coo.col, pre_spike, post_trace))

            mat = mat + u.math.outer(pre_spike, post_trace)
            mat = u.math.where(mask, mat, 0. * mat_unit)

            assert u.math.allclose(coo.todense(), mat)
            jax.block_until_ready((mat, pre_spike, post_trace))

        if mat_unit.has_same_dim(trace_unit):
            run()
        else:
            with pytest.raises(u.UnitMismatchError):
                run()

    @pytest.mark.parametrize('w_in', [None, 0.1])
    @pytest.mark.parametrize('w_max', [None, 0.5])
    def test_coo_on_pre_v2(self, w_in, w_max):
        n_pre = 100
        n_post = 100
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        pre_spike = brainstate.random.random((n_pre,)) < 0.1
        post_trace = brainstate.random.random((n_post,))

        coo = brainevent.COO.fromdense(mat)
        coo = coo.with_data(update_coo_on_binary_pre(coo.data, coo.row, coo.col, pre_spike, post_trace, w_in, w_max))

        mat = mat + jnp.outer(pre_spike, post_trace)
        mat = jnp.clip(mat, w_in, w_max)
        mat = jnp.where(mask, mat, 0.)

        assert jnp.allclose(coo.todense(), mat)
        jax.block_until_ready((mat, pre_spike, post_trace))


class Test_coo_on_post:

    def test_coo_on_post_v1(self):
        n_pre = 100
        n_post = 100
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        pre_trace = brainstate.random.random((n_pre,))
        post_spike = brainstate.random.random((n_post,)) < 0.1

        coo = brainevent.COO.fromdense(mat)
        coo = coo.with_data(update_coo_on_binary_post(coo.data, coo.row, coo.col, pre_trace, post_spike))

        mat = mat + jnp.outer(pre_trace, post_spike)
        mat = jnp.where(mask, mat, 0.)

        assert jnp.allclose(coo.todense(), mat)
        jax.block_until_ready((mat, pre_trace, post_spike))

    @pytest.mark.parametrize('mat_unit', [u.mV, u.ms])
    @pytest.mark.parametrize('trace_unit', [u.mV, u.ms])
    def test_coo_on_post_unit(self, mat_unit, trace_unit):
        def run():
            n_pre = 100
            n_post = 100
            mat = brainstate.random.random((n_pre, n_post))
            mask = mat < 0.5
            mat = jnp.where(mask, mat, 0.) * mat_unit
            pre_trace = brainstate.random.random((n_pre,)) * trace_unit
            post_spike = brainstate.random.random((n_post,)) < 0.1

            coo = brainevent.COO.fromdense(mat)
            coo = coo.with_data(update_coo_on_binary_post(coo.data, coo.row, coo.col, pre_trace, post_spike))

            mat = mat + u.math.outer(pre_trace, post_spike)
            mat = u.math.where(mask, mat, 0. * mat_unit)

            assert u.math.allclose(coo.todense(), mat)
            jax.block_until_ready((mat, pre_trace, post_spike))

        if mat_unit.has_same_dim(trace_unit):
            run()
        else:
            with pytest.raises(u.UnitMismatchError):
                run()

    @pytest.mark.parametrize('w_in', [None, 0.1])
    @pytest.mark.parametrize('w_max', [None, 0.5])
    def test_coo_on_post_v2(self, w_in, w_max):
        n_pre = 100
        n_post = 100
        mat = brainstate.random.random((n_pre, n_post))
        mask = mat < 0.5
        mat = jnp.where(mask, mat, 0.)
        pre_trace = brainstate.random.random((n_pre,))
        post_spike = brainstate.random.random((n_post,)) < 0.1

        coo = brainevent.COO.fromdense(mat)
        coo = coo.with_data(update_coo_on_binary_post(coo.data, coo.row, coo.col, pre_trace, post_spike, w_in, w_max))

        mat = mat + jnp.outer(pre_trace, post_spike)
        mat = jnp.clip(mat, w_in, w_max)
        mat = jnp.where(mask, mat, 0.)

        assert jnp.allclose(coo.todense(), mat)
        jax.block_until_ready((mat, pre_trace, post_spike))


@pytest.mark.skipif(PLATFORM != 'gpu', reason='GPU backend parity tests require CUDA GPU.')
class Test_coo_gpu_backend_parity:
    @staticmethod
    def _call_pre_backend(weight, pre_ids, post_ids, pre_spike, post_trace, backend):
        return update_coo_on_binary_pre_p(
            weight, pre_ids, post_ids, pre_spike, post_trace,
            outs=[jax.ShapeDtypeStruct(weight.shape, weight.dtype)],
            weight_info=jax.ShapeDtypeStruct(weight.shape, weight.dtype),
            pre_ids_info=jax.ShapeDtypeStruct(pre_ids.shape, pre_ids.dtype),
            post_ids_info=jax.ShapeDtypeStruct(post_ids.shape, post_ids.dtype),
            spike_info=jax.ShapeDtypeStruct(pre_spike.shape, pre_spike.dtype),
            trace_info=jax.ShapeDtypeStruct(post_trace.shape, post_trace.dtype),
            backend=backend,
        )[0]

    @staticmethod
    def _call_post_backend(weight, pre_ids, post_ids, pre_trace, post_spike, backend):
        return update_coo_on_binary_post_p(
            weight, pre_ids, post_ids, pre_trace, post_spike,
            outs=[jax.ShapeDtypeStruct(weight.shape, weight.dtype)],
            weight_info=jax.ShapeDtypeStruct(weight.shape, weight.dtype),
            pre_ids_info=jax.ShapeDtypeStruct(pre_ids.shape, pre_ids.dtype),
            post_ids_info=jax.ShapeDtypeStruct(post_ids.shape, post_ids.dtype),
            trace_info=jax.ShapeDtypeStruct(pre_trace.shape, pre_trace.dtype),
            spike_info=jax.ShapeDtypeStruct(post_spike.shape, post_spike.dtype),
            backend=backend,
        )[0]

    @staticmethod
    def _make_ids(n_syn, n_pre, n_post, seed):
        rng = np.random.default_rng(seed)
        if n_syn == 0:
            return jnp.empty((0,), dtype=jnp.int32), jnp.empty((0,), dtype=jnp.int32)
        # Skew toward a hotspot to stress duplicate gathers.
        pre_ids = rng.integers(0, n_pre, size=n_syn, dtype=np.int32)
        post_ids = rng.integers(0, n_post, size=n_syn, dtype=np.int32)
        hot = max(1, n_syn // 8)
        pre_ids[:hot] = pre_ids[0]
        post_ids[:hot] = post_ids[0]
        return jnp.asarray(pre_ids), jnp.asarray(post_ids)

    @pytest.mark.parametrize('backend', [b for b in ('warp', 'pallas') if b in PRE_BACKENDS])
    @pytest.mark.parametrize('dtype', [jnp.float32, jnp.float16])
    @pytest.mark.parametrize('bool_spike', [True, False])
    @pytest.mark.parametrize('n_syn', [0, 33, 4097])
    def test_pre_backend_matches_reference(self, backend, dtype, bool_spike, n_syn):
        n_pre, n_post = 257, 193
        rng = np.random.default_rng(123)
        pre_ids, post_ids = self._make_ids(n_syn, n_pre, n_post, seed=321)
        weight = jnp.asarray(rng.standard_normal(n_syn), dtype=dtype)
        post_trace = jnp.asarray(rng.standard_normal(n_post), dtype=dtype)
        if bool_spike:
            pre_spike = jnp.asarray(rng.random(n_pre) > 0.6, dtype=jnp.bool_)
            spike_mask = pre_spike[pre_ids]
        else:
            pre_spike = jnp.asarray(rng.standard_normal(n_pre), dtype=dtype)
            spike_mask = pre_spike[pre_ids] != 0.

        f = jax.jit(lambda: self._call_pre_backend(weight, pre_ids, post_ids, pre_spike, post_trace, backend))
        out = f()
        ref = weight + jnp.where(spike_mask, post_trace[post_ids], 0)

        rtol, atol = (5e-3, 5e-3) if dtype == jnp.float16 else (1e-5, 1e-5)
        assert jnp.allclose(out, ref, rtol=rtol, atol=atol)
        jax.block_until_ready((out, ref))

    @pytest.mark.parametrize('backend', [b for b in ('warp', 'pallas') if b in POST_BACKENDS])
    @pytest.mark.parametrize('dtype', [jnp.float32, jnp.float16])
    @pytest.mark.parametrize('bool_spike', [True, False])
    @pytest.mark.parametrize('n_syn', [0, 33, 4097])
    def test_post_backend_matches_reference(self, backend, dtype, bool_spike, n_syn):
        n_pre, n_post = 257, 193
        rng = np.random.default_rng(456)
        pre_ids, post_ids = self._make_ids(n_syn, n_pre, n_post, seed=654)
        weight = jnp.asarray(rng.standard_normal(n_syn), dtype=dtype)
        pre_trace = jnp.asarray(rng.standard_normal(n_pre), dtype=dtype)
        if bool_spike:
            post_spike = jnp.asarray(rng.random(n_post) > 0.6, dtype=jnp.bool_)
            spike_mask = post_spike[post_ids]
        else:
            post_spike = jnp.asarray(rng.standard_normal(n_post), dtype=dtype)
            spike_mask = post_spike[post_ids] != 0.

        f = jax.jit(lambda: self._call_post_backend(weight, pre_ids, post_ids, pre_trace, post_spike, backend))
        out = f()
        ref = weight + jnp.where(spike_mask, pre_trace[pre_ids], 0)

        rtol, atol = (5e-3, 5e-3) if dtype == jnp.float16 else (1e-5, 1e-5)
        assert jnp.allclose(out, ref, rtol=rtol, atol=atol)
        jax.block_until_ready((out, ref))

    def test_default_backend_routes_bf16_to_supported_kernel(self):
        n_pre, n_post, n_syn = 127, 111, 4099
        rng = np.random.default_rng(777)
        pre_ids, post_ids = self._make_ids(n_syn, n_pre, n_post, seed=778)
        weight = jnp.asarray(rng.standard_normal(n_syn), dtype=jnp.bfloat16)
        pre_trace = jnp.asarray(rng.standard_normal(n_pre), dtype=jnp.bfloat16)
        post_trace = jnp.asarray(rng.standard_normal(n_post), dtype=jnp.bfloat16)
        pre_spike = jnp.asarray(rng.random(n_pre) > 0.5, dtype=jnp.bool_)
        post_spike = jnp.asarray(rng.random(n_post) > 0.5, dtype=jnp.bool_)

        out_pre = update_coo_on_binary_pre(weight, pre_ids, post_ids, pre_spike, post_trace, backend='pallas')
        out_post = update_coo_on_binary_post(weight, pre_ids, post_ids, pre_trace, post_spike, backend='pallas')

        ref_pre = weight + jnp.where(pre_spike[pre_ids], post_trace[post_ids], 0)
        ref_post = weight + jnp.where(post_spike[post_ids], pre_trace[pre_ids], 0)

        assert jnp.allclose(out_pre, ref_pre, rtol=2e-2, atol=2e-2)
        assert jnp.allclose(out_post, ref_post, rtol=2e-2, atol=2e-2)
        jax.block_until_ready((out_pre, out_post, ref_pre, ref_post))

    @pytest.mark.skipif(not jax.config.read('jax_enable_x64'), reason='Requires jax_enable_x64=True')
    def test_pallas_int64_indices_parity(self):
        if 'pallas' not in PRE_BACKENDS or 'pallas' not in POST_BACKENDS:
            pytest.skip('Pallas backend unavailable for COO plasticity kernels.')

        n_pre, n_post, n_syn = 257, 193, 2048
        rng = np.random.default_rng(999)
        pre_ids = jnp.asarray(rng.integers(0, n_pre, size=n_syn, dtype=np.int64))
        post_ids = jnp.asarray(rng.integers(0, n_post, size=n_syn, dtype=np.int64))
        weight = jnp.asarray(rng.standard_normal(n_syn), dtype=jnp.float32)
        pre_trace = jnp.asarray(rng.standard_normal(n_pre), dtype=jnp.float32)
        post_trace = jnp.asarray(rng.standard_normal(n_post), dtype=jnp.float32)
        pre_spike = jnp.asarray(rng.standard_normal(n_pre), dtype=jnp.float32)
        post_spike = jnp.asarray(rng.standard_normal(n_post), dtype=jnp.float32)

        out_pre = jax.jit(
            lambda: self._call_pre_backend(weight, pre_ids, post_ids, pre_spike, post_trace, backend='pallas')
        )()
        out_post = jax.jit(
            lambda: self._call_post_backend(weight, pre_ids, post_ids, pre_trace, post_spike, backend='pallas')
        )()

        ref_pre = weight + jnp.where(pre_spike[pre_ids] != 0., post_trace[post_ids], 0)
        ref_post = weight + jnp.where(post_spike[post_ids] != 0., pre_trace[pre_ids], 0)

        assert jnp.allclose(out_pre, ref_pre, rtol=1e-5, atol=1e-5)
        assert jnp.allclose(out_post, ref_post, rtol=1e-5, atol=1e-5)
        jax.block_until_ready((out_pre, out_post, ref_pre, ref_post))



# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import brainstate
import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._dense.indexed_binary import (
    indexed_binary_densemv_p_call,
    indexed_binary_densemv_p,
    indexed_binary_densemm_p_call,
    indexed_binary_densemm_p,
)

platform = jax.default_backend()
MV_IMPLEMENTATIONS = tuple(indexed_binary_densemv_p.available_backends(platform))
MM_IMPLEMENTATIONS = tuple(indexed_binary_densemm_p.available_backends(platform))


@pytest.mark.skipif(not MV_IMPLEMENTATIONS, reason=f'No indexed_binary_densemv implementation on platform={platform}')
@pytest.mark.parametrize('implementation', MV_IMPLEMENTATIONS)
def test_indexed_binary_densemv_transpose_true(implementation):
    """transpose=True: weights[k,n], select rows by indices -> out[n] (old indexed_bdvm)"""
    n_input, n_output = 8, 6
    weights = brainstate.random.randn(n_input, n_output)
    spikes = jnp.ones((n_input,), dtype=jnp.float32)
    indices = jnp.asarray([1, 7, 99, -3], dtype=jnp.int32)
    count = jnp.asarray([16], dtype=jnp.int32)
    result = indexed_binary_densemv_p_call(spikes, indices, count, weights, transpose=True, backend=implementation)[0]

    weights_np = np.asarray(weights)
    indices_np = np.asarray(indices)
    valid_indices = indices_np[(indices_np >= 0) & (indices_np < n_input)]
    expected = jnp.asarray(weights_np[valid_indices].sum(axis=0), dtype=weights.dtype)
    assert u.math.allclose(result, expected, atol=1e-5, rtol=1e-5)
    jax.block_until_ready((result, expected))


@pytest.mark.skipif(not MV_IMPLEMENTATIONS, reason=f'No indexed_binary_densemv implementation on platform={platform}')
@pytest.mark.parametrize('implementation', MV_IMPLEMENTATIONS)
def test_indexed_binary_densemv_transpose_false(implementation):
    """transpose=False: weights[m,k], select columns by indices -> out[m] (old indexed_dbmv)"""
    n_input, n_output = 6, 8
    weights = brainstate.random.randn(n_input, n_output)
    spikes = jnp.ones((n_output,), dtype=jnp.float32)
    indices = jnp.asarray([1, 7, 99, -3], dtype=jnp.int32)
    count = jnp.asarray([16], dtype=jnp.int32)
    result = indexed_binary_densemv_p_call(spikes, indices, count, weights, transpose=False, backend=implementation)[0]

    weights_np = np.asarray(weights)
    indices_np = np.asarray(indices)
    valid_indices = indices_np[(indices_np >= 0) & (indices_np < n_output)]
    expected = jnp.asarray(weights_np[:, valid_indices].sum(axis=1), dtype=weights.dtype)
    assert u.math.allclose(result, expected, atol=1e-5, rtol=1e-5)
    jax.block_until_ready((result, expected))


@pytest.mark.skipif(not MM_IMPLEMENTATIONS, reason=f'No indexed_binary_densemm implementation on platform={platform}')
@pytest.mark.parametrize('implementation', MM_IMPLEMENTATIONS)
def test_indexed_binary_densemm_transpose_true(implementation):
    """transpose=True: weights[k,n], select rows by indices -> out[batch, n] (old indexed_bdmm)"""
    batch, n_input, n_output = 2, 9, 5
    weights = brainstate.random.randn(n_input, n_output)
    spikes = jnp.ones((batch, n_input), dtype=jnp.float32)
    indices = jnp.asarray([[0, 3, 99, -1], [8, 1, 4, 7]], dtype=jnp.int32)
    count = jnp.asarray([4, 12], dtype=jnp.int32)
    result = indexed_binary_densemm_p_call(spikes, indices, count, weights, transpose=True, backend=implementation)[0]

    weights_np = np.asarray(weights)
    indices_np = np.asarray(indices)
    expected_rows = []
    for b in range(batch):
        idx = indices_np[b]
        valid_indices = idx[(idx >= 0) & (idx < n_input)]
        expected_rows.append(weights_np[valid_indices].sum(axis=0))
    expected = jnp.asarray(np.stack(expected_rows, axis=0), dtype=weights.dtype)

    assert u.math.allclose(result, expected, atol=1e-5, rtol=1e-5)
    jax.block_until_ready((result, expected))


@pytest.mark.skipif(not MM_IMPLEMENTATIONS, reason=f'No indexed_binary_densemm implementation on platform={platform}')
@pytest.mark.parametrize('implementation', MM_IMPLEMENTATIONS)
def test_indexed_binary_densemm_transpose_false(implementation):
    """transpose=False: weights[m,k], select columns by indices -> out[batch, m] (old indexed_dbmm)"""
    batch, n_input, n_output = 2, 5, 9
    weights = brainstate.random.randn(n_input, n_output)
    spikes = jnp.ones((batch, n_output), dtype=jnp.float32)
    indices = jnp.asarray([[0, 3, 99, -1], [8, 1, 4, 7]], dtype=jnp.int32)
    count = jnp.asarray([4, 12], dtype=jnp.int32)
    result = indexed_binary_densemm_p_call(spikes, indices, count, weights, transpose=False, backend=implementation)[0]

    weights_np = np.asarray(weights)
    indices_np = np.asarray(indices)
    expected_rows = []
    for b in range(batch):
        idx = indices_np[b]
        valid_indices = idx[(idx >= 0) & (idx < n_output)]
        expected_rows.append(weights_np[:, valid_indices].sum(axis=1))
    expected = jnp.asarray(np.stack(expected_rows, axis=0), dtype=weights.dtype)

    assert u.math.allclose(result, expected, atol=1e-5, rtol=1e-5)
    jax.block_until_ready((result, expected))



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._coo.test_util import vector_coo, coo_vector, matrix_coo, coo_matrix


def _dense_from_coo(row, col, data, shape):
    row = np.asarray(row, dtype=np.int64)
    col = np.asarray(col, dtype=np.int64)
    data = np.asarray(data)
    dense = np.zeros(shape, dtype=data.dtype)
    for r, c, d in zip(row, col, data):
        dense[r, c] += d
    return dense


def _sample_coo(shape, nnz, *, seed=0, allow_duplicates=True):
    rng = np.random.default_rng(seed)
    m, n = shape
    if nnz == 0:
        return np.array([], dtype=np.int32), np.array([], dtype=np.int32)
    if allow_duplicates:
        row = rng.integers(0, m, size=nnz, dtype=np.int32)
        col = rng.integers(0, n, size=nnz, dtype=np.int32)
        return row, col
    flat = rng.choice(m * n, size=nnz, replace=False)
    row = (flat // n).astype(np.int32)
    col = (flat % n).astype(np.int32)
    return row, col


@pytest.mark.parametrize("homo_w", [True, False])
@pytest.mark.parametrize("allow_duplicates", [True, False])
def test_vector_coo_matches_dense(homo_w, allow_duplicates):
    shape = (6, 5)
    row, col = _sample_coo(shape, 12, seed=1, allow_duplicates=allow_duplicates)
    rng = np.random.default_rng(2)
    x = rng.normal(size=shape[0]).astype(np.float32)

    if homo_w:
        w = np.float32(1.25)
        data = np.full(row.shape, w, dtype=np.float32)
    else:
        data = rng.normal(size=row.shape[0]).astype(np.float32)
        w = data

    dense = _dense_from_coo(row, col, data, shape)
    y_ref = x @ dense
    y = vector_coo(jnp.asarray(x), w, row, col, shape)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((y,))


@pytest.mark.parametrize("homo_w", [True, False])
@pytest.mark.parametrize("allow_duplicates", [True, False])
def test_coo_vector_matches_dense(homo_w, allow_duplicates):
    shape = (6, 5)
    row, col = _sample_coo(shape, 12, seed=3, allow_duplicates=allow_duplicates)
    rng = np.random.default_rng(4)
    v = rng.normal(size=shape[1]).astype(np.float32)

    if homo_w:
        w = np.float32(0.5)
        data = np.full(row.shape, w, dtype=np.float32)
    else:
        data = rng.normal(size=row.shape[0]).astype(np.float32)
        w = data

    dense = _dense_from_coo(row, col, data, shape)
    y_ref = dense @ v
    y = coo_vector(jnp.asarray(v), w, row, col, shape)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((y,))


@pytest.mark.parametrize("homo_w", [True, False])
@pytest.mark.parametrize("allow_duplicates", [True, False])
def test_matrix_coo_matches_dense(homo_w, allow_duplicates):
    shape = (5, 4)
    row, col = _sample_coo(shape, 10, seed=5, allow_duplicates=allow_duplicates)
    rng = np.random.default_rng(6)
    xs = rng.normal(size=(3, shape[0])).astype(np.float32)

    if homo_w:
        w = np.float32(2.0)
        data = np.full(row.shape, w, dtype=np.float32)
    else:
        data = rng.normal(size=row.shape[0]).astype(np.float32)
        w = data

    dense = _dense_from_coo(row, col, data, shape)
    y_ref = xs @ dense
    y = matrix_coo(jnp.asarray(xs), w, row, col, shape)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((y,))


@pytest.mark.parametrize("homo_w", [True, False])
@pytest.mark.parametrize("allow_duplicates", [True, False])
def test_coo_matrix_matches_dense(homo_w, allow_duplicates):
    shape = (5, 4)
    row, col = _sample_coo(shape, 10, seed=7, allow_duplicates=allow_duplicates)
    rng = np.random.default_rng(8)
    xs = rng.normal(size=(shape[1], 3)).astype(np.float32)

    if homo_w:
        w = np.float32(1.75)
        data = np.full(row.shape, w, dtype=np.float32)
    else:
        data = rng.normal(size=row.shape[0]).astype(np.float32)
        w = data

    dense = _dense_from_coo(row, col, data, shape)
    y_ref = dense @ xs
    y = coo_matrix(jnp.asarray(xs), w, row, col, shape)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((y,))


def test_empty_coo_outputs_zeros():
    shape = (4, 3)
    row = np.array([], dtype=np.int32)
    col = np.array([], dtype=np.int32)

    x = jnp.arange(shape[0], dtype=jnp.float32)
    v = jnp.arange(shape[1], dtype=jnp.float32)
    xs_left = jnp.arange(2 * shape[0], dtype=jnp.float32).reshape(2, shape[0])
    xs_right = jnp.arange(shape[1] * 2, dtype=jnp.float32).reshape(shape[1], 2)

    out_vector = vector_coo(x, 1.0, row, col, shape)
    assert out_vector.shape == (shape[1],)
    assert jnp.all(out_vector == 0)

    out_coo_vector = coo_vector(v, jnp.array([], dtype=jnp.float32), row, col, shape)
    assert out_coo_vector.shape == (shape[0],)
    assert jnp.all(out_coo_vector == 0)

    out_matrix = matrix_coo(xs_left, 1.0, row, col, shape)
    assert out_matrix.shape == (xs_left.shape[0], shape[1])
    assert jnp.all(out_matrix == 0)

    out_coo_matrix = coo_matrix(xs_right, jnp.array([], dtype=jnp.float32), row, col, shape)
    assert out_coo_matrix.shape == (shape[0], xs_right.shape[1])
    assert jnp.all(out_coo_matrix == 0)
    jax.block_until_ready((x, v, xs_left, xs_right, out_vector, out_coo_vector, out_matrix, out_coo_matrix))


def test_matrix_coo_jit_matches_dense():
    shape = (6, 5)
    row, col = _sample_coo(shape, 8, seed=9, allow_duplicates=False)
    xs = jnp.arange(18, dtype=jnp.float32).reshape(3, shape[0]) / 10.0
    w = jnp.linspace(0.1, 0.8, num=row.shape[0], dtype=jnp.float32)

    dense = jnp.zeros(shape, dtype=jnp.float32).at[row, col].add(w)
    y_ref = xs @ dense

    f = jax.jit(lambda x: matrix_coo(x, w, row, col, shape))
    y = f(xs)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((xs, w, dense, y_ref, y))


def test_coo_matrix_jit_matches_dense():
    shape = (6, 5)
    row, col = _sample_coo(shape, 8, seed=10, allow_duplicates=False)
    xs = jnp.arange(15, dtype=jnp.float32).reshape(shape[1], 3) / 7.0
    w = jnp.linspace(0.2, 1.0, num=row.shape[0], dtype=jnp.float32)

    dense = jnp.zeros(shape, dtype=jnp.float32).at[row, col].add(w)
    y_ref = dense @ xs

    f = jax.jit(lambda x: coo_matrix(x, w, row, col, shape))
    y = f(xs)
    assert jnp.allclose(y, y_ref, rtol=1e-5, atol=1e-5)
    jax.block_until_ready((xs, w, dense, y_ref, y))


def test_matrix_coo_grad_matches_dense():
    shape = (4, 3)
    row, col = _sample_coo(shape, 6, seed=11, allow_duplicates=False)
    xs = jnp.arange(8, dtype=jnp.float32).reshape(2, shape[0]) / 5.0
    w = jnp.linspace(0.1, 0.6, num=row.shape[0], dtype=jnp.float32)

    def f(xs, w):
        return matrix_coo(xs, w, row, col, shape).sum()

    def f_ref(xs, w):
        dense = jnp.zeros(shape, dtype=xs.dtype).at[row, col].add(w)
        return (xs @ dense).sum()

    g = jax.grad(f, argnums=(0, 1))(xs, w)
    g_ref = jax.grad(f_ref, argnums=(0, 1))(xs, w)

    assert jnp.allclose(g[0], g_ref[0], rtol=1e-5, atol=1e-5)
    assert jnp.allclose(g[1], g_ref[1], rtol=1e-5, atol=1e-5)
    jax.block_until_ready((xs, w, g, g_ref))


def test_coo_matrix_grad_matches_dense():
    shape = (4, 3)
    row, col = _sample_coo(shape, 6, seed=12, allow_duplicates=False)
    xs = jnp.arange(9, dtype=jnp.float32).reshape(shape[1], 3) / 4.0
    w = jnp.linspace(0.2, 0.7, num=row.shape[0], dtype=jnp.float32)

    def f(xs, w):
        return coo_matrix(xs, w, row, col, shape).sum()

    def f_ref(xs, w):
        dense = jnp.zeros(shape, dtype=xs.dtype).at[row, col].add(w)
        return (dense @ xs).sum()

    g = jax.grad(f, argnums=(0, 1))(xs, w)
    g_ref = jax.grad(f_ref, argnums=(0, 1))(xs, w)

    assert jnp.allclose(g[0], g_ref[0], rtol=1e-5, atol=1e-5)
    assert jnp.allclose(g[1], g_ref[1], rtol=1e-5, atol=1e-5)
    jax.block_until_ready((xs, w, g, g_ref))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# -*- coding: utf-8 -*-

from typing import Optional, Sequence

import brainunit as u
import jax
import numpy as np
from jax import numpy as jnp
from jax.interpreters import ad

from brainevent._data import _initialize_seed, _initialize_conn_length
from brainevent._misc import generate_block_dim, namescope
from brainevent._numba_random import get_numba_lfsr_seed, get_numba_lfsr_random_integers, get_numba_lfsr_normal
from brainevent._op import XLACustomKernel, jaxinfo_to_warpinfo, numba_kernel, general_batching_rule
from brainevent._op.benchmark import BenchmarkConfig
from brainevent._pallas_random import get_pallas_lfsr_rng_class
from brainevent._typing import Data, MatrixShape
from .float import jitnmv_p_call, jitnmm_p_call

__all__ = [
    "binary_jitnmv",
    "binary_jitnmv_p",
    "binary_jitnmm",
    "binary_jitnmm_p",
]


@namescope(static_argnames=("shape", "transpose", "corder"))
def binary_jitnmv(
    w_loc: Data,
    w_scale: Data,
    prob: float,
    vector: Data,
    seed: Optional[int] = None,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    corder: bool = True,
    backend: Optional[str] = None,
) -> Data:
    """
    Event-driven matrix-vector multiplication with a JIT normal-distributed connectivity matrix.

    Computes ``M @ v`` where ``M`` is a sparse matrix whose non-zero entries are
    drawn from a normal distribution with parameters ``w_loc`` (mean) and ``w_scale``
    (standard deviation), and ``v`` is a binary event vector. Only positions where
    ``v`` is active (True or > 0) contribute to the output, making this operation
    event-driven and efficient for sparse neural activity patterns.

    Parameters
    ----------
    w_loc : Data
        Location (mean) parameter of the normal distribution for the matrix
        weights. Scalar or 1-element array, optionally with physical units.
    w_scale : Data
        Scale (standard deviation) parameter of the normal distribution for the
        matrix weights. Must have the same physical dimension as ``w_loc``.
    prob : float
        Connection probability in the range ``[0, 1]``. Controls the sparsity of
        the generated connectivity matrix.
    vector : Data
        The binary event vector to multiply with. Elements are treated as events
        (True/False or >0/<=0). Shape must be compatible with ``shape``.
    seed : int, optional
        Random seed for reproducible matrix generation. If None, a random seed
        is generated at compile time.
    shape : MatrixShape
        Shape of the implicit connectivity matrix as ``(rows, cols)``.
    transpose : bool, optional
        If True, compute ``M.T @ v`` instead of ``M @ v``. Default is False.
    corder : bool, optional
        Memory layout order for kernel dispatch. True for C-order (row-major),
        False for Fortran-order (column-major). Default is True.
    backend : str, optional
        Compute backend to use (``'numba'``, ``'warp'``, ``'pallas'``, or None
        for automatic selection).

    Returns
    -------
    Data
        The result vector of the matrix-vector product. If the inputs carry
        physical units, the output will have units equal to the product of the
        weight units and the vector units.

    Raises
    ------
    brainunit.DimensionMismatchError
        If ``w_loc`` and ``w_scale`` do not have the same physical dimension.

    See Also
    --------
    binary_jitnmm : Event-driven matrix-matrix multiplication variant.
    jitnmv : Float (non-event) matrix-vector multiplication with normal weights.

    Notes
    -----
    The connectivity matrix ``W`` is never materialized in memory. Instead, the
    pseudo-random structure is regenerated on-the-fly using the ``seed`` and
    ``prob`` parameters, following the same PRNG sequence as ``jitn`` to ensure
    consistency with ``todense()``.

    The implicit weight matrix has entries:

    ``W[i, j] = Normal(w_loc, w_scale) * Bernoulli(prob)``

    where ``Normal(w_loc, w_scale)`` is an independent draw for each non-zero
    position, and ``Bernoulli(prob)`` is 1 with probability ``prob`` and 0
    otherwise. The connection mask is determined by a deterministic hash of
    ``(seed, i, j)``.

    The event-driven matrix-vector product computes:

    ``y[i] = sum_{j in C(i)} N_ij * spike[j]``

    where ``C(i) = {j : Bernoulli_ij = 1}`` is the set of connected
    pre-synaptic indices for post-synaptic neuron ``i``, ``N_ij ~ Normal(w_loc,
    w_scale)`` is the connection weight, and ``spike[j]`` is treated as a
    binary event (True/False or >0/<=0). Equivalently:

    ``y[i] = sum_{j in C(i) : spike[j]=1} N_ij``

    The connection length parameter ``clen = 2 / prob`` controls the average
    stride between non-zero entries.

    This operation supports automatic differentiation (JVP and transpose rules)
    for ``w_loc``, ``w_scale``, and ``vector``. Batching over the vector
    dimension is promoted to ``binary_jitnmm``.

    Examples
    --------

    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.binary import binary_jitnmv
        >>> w_loc = jnp.array([1.0])
        >>> w_scale = jnp.array([0.1])
        >>> events = jnp.array([True, False, True, True, False])
        >>> result = binary_jitnmv(w_loc, w_scale, 0.5, events, seed=42,
        ...                        shape=(3, 5))
    """
    u.fail_for_dimension_mismatch(w_loc, w_scale, "w_loc and w_scale must have the same dimension.")
    seed = _initialize_seed(seed)
    w_loc, unitd = u.split_mantissa_unit(w_loc)
    w_scale = u.Quantity(w_scale).to(unitd).mantissa
    vector, unitv = u.split_mantissa_unit(vector)
    clen = _initialize_conn_length(prob)
    res = binary_jitnmv_p_call(
        w_loc,
        w_scale,
        clen,
        vector,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * unitd * unitv)


def binary_jitnmm(
    w_loc: Data,
    w_scale: Data,
    prob: float,
    B: Data,
    seed: Optional[int] = None,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    corder: bool = True,
    backend: Optional[str] = None,
) -> Data:
    """
    Event-driven matrix-matrix multiplication with a JIT normal-distributed connectivity matrix.

    Computes ``M @ B`` where ``M`` is a sparse matrix whose non-zero entries are
    drawn from a normal distribution with parameters ``w_loc`` (mean) and ``w_scale``
    (standard deviation), and ``B`` is a 2-D binary event matrix. Only positions
    where ``B`` elements are active (True or > 0) contribute to the output.

    Parameters
    ----------
    w_loc : Data
        Location (mean) parameter of the normal distribution for the matrix
        weights. Scalar or 1-element array, optionally with physical units.
    w_scale : Data
        Scale (standard deviation) parameter of the normal distribution for the
        matrix weights. Must have the same physical dimension as ``w_loc``.
    prob : float
        Connection probability in the range ``[0, 1]``. Controls the sparsity of
        the generated connectivity matrix.
    B : Data
        The binary event matrix to multiply with, shape ``(k, n)``. Elements are
        treated as events (True/False or >0/<=0).
    seed : int, optional
        Random seed for reproducible matrix generation. If None, a random seed
        is generated at compile time.
    shape : MatrixShape
        Shape of the implicit connectivity matrix as ``(rows, cols)``.
    transpose : bool, optional
        If True, compute ``M.T @ B`` instead of ``M @ B``. Default is False.
    corder : bool, optional
        Memory layout order for kernel dispatch. True for C-order (row-major),
        False for Fortran-order (column-major). Default is True.
    backend : str, optional
        Compute backend to use (``'numba'``, ``'warp'``, ``'pallas'``, or None
        for automatic selection).

    Returns
    -------
    Data
        The result matrix of the matrix-matrix product with shape
        ``(shape[0], B.shape[1])`` (or ``(shape[1], B.shape[1])`` if transposed).
        If the inputs carry physical units, the output will have units equal to
        the product of the weight units and the ``B`` units.

    Raises
    ------
    brainunit.DimensionMismatchError
        If ``w_loc`` and ``w_scale`` do not have the same physical dimension.

    See Also
    --------
    binary_jitnmv : Event-driven matrix-vector multiplication variant.
    jitnmm : Float (non-event) matrix-matrix multiplication with normal weights.

    Notes
    -----
    The connectivity matrix ``W`` is never materialized in memory. The
    pseudo-random structure is regenerated on-the-fly using the ``seed`` and
    ``prob`` parameters, matching the PRNG sequence used by ``jitn``.

    The implicit weight matrix has entries:

    ``W[i, j] = Normal(w_loc, w_scale) * Bernoulli(prob)``

    where ``Normal(w_loc, w_scale)`` is an independent draw for each non-zero
    position, and ``Bernoulli(prob)`` is 1 with probability ``prob`` and 0
    otherwise.

    The event-driven matrix-matrix product computes:

    ``Y[i, k] = sum_{j in C(i)} N_ij * spike[j, k]``

    where ``C(i) = {j : Bernoulli_ij = 1}`` is the set of connected
    pre-synaptic indices for post-synaptic neuron ``i``, ``N_ij ~ Normal(w_loc,
    w_scale)`` is the connection weight, and ``spike[j, k]`` is treated as a
    binary event. Each column ``k`` of ``B`` is processed independently.

    The connection length parameter ``clen = 2 / prob`` controls the average
    stride between non-zero entries.

    This operation supports automatic differentiation (JVP and transpose rules)
    for ``w_loc``, ``w_scale``, and ``B``. Batching over the ``B`` dimension
    is supported along axes 0, 1, and 2.

    Examples
    --------

    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.binary import binary_jitnmm
        >>> w_loc = jnp.array([1.0])
        >>> w_scale = jnp.array([0.1])
        >>> B = jnp.array([[True, False], [False, True], [True, True],
        ...                [False, False], [True, False]])
        >>> result = binary_jitnmm(w_loc, w_scale, 0.5, B, seed=42,
        ...                        shape=(3, 5))
    """
    u.fail_for_dimension_mismatch(w_loc, w_scale, "w_loc and w_scale must have the same dimension.")
    seed = _initialize_seed(seed)
    w_loc, unitd = u.split_mantissa_unit(w_loc)
    w_scale = u.Quantity(w_scale).to(unitd).mantissa
    B, unitB = u.split_mantissa_unit(B)
    clen = _initialize_conn_length(prob)
    res = binary_jitnmm_p_call(
        w_loc,
        w_scale,
        clen,
        B,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * unitd * unitB)


def _jitc_mv_normal_numba_kernel_generator(
    corder: bool,
    vector_info: jax.ShapeDtypeStruct,
    **kwargs
):
    r"""Generate the CPU kernel for the :func:`_jitc_matvec_normal` operation.
    """
    import numba
    _lfsr_seed = get_numba_lfsr_seed()
    _lfsr_random_integers = get_numba_lfsr_random_integers()
    _lfsr_normal = get_numba_lfsr_normal()

    if corder:
        # This means that the for loop is parallelized along the dimension of the output vector: ``post.shape[0]``.
        if vector_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, vector, seed, posts):
                posts[:] = 0.
                n_col = posts.shape[0]
                n_row = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                for i_col in range(n_col):
                    state = _lfsr_seed(seed0 + i_col * n_row)
                    i_row = _lfsr_random_integers(state, 0, clen0 - 1)
                    out = np.asarray(0., dtype=posts.dtype)
                    while i_row < n_row:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        if vector[i_row]:
                            out += w
                        i_row += _lfsr_random_integers(state, 1, clen0 - 1)
                    posts[i_col] = out
        else:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, vector, seed, posts):
                posts[:] = 0.
                n_col = posts.shape[0]
                n_row = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                for i_col in range(n_col):
                    state = _lfsr_seed(seed0 + i_col * n_row)
                    i_row = _lfsr_random_integers(state, 0, clen0 - 1)
                    out = np.asarray(0., dtype=posts.dtype)
                    while i_row < n_row:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        if vector[i_row] > 0.:
                            out += w
                        i_row += _lfsr_random_integers(state, 1, clen0 - 1)
                    posts[i_col] = out

    else:
        if vector_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, vector, seed, posts):
                posts[:] = 0.
                num_col = posts.shape[0]
                num_row = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                for i_row in range(num_row):
                    if vector[i_row]:
                        state = _lfsr_seed(seed0 + i_row * num_col)
                        i_col = _lfsr_random_integers(state, 0, clen0 - 1)
                        while i_col < num_col:
                            w = _lfsr_normal(state, w_loc0, w_scale0)
                            posts[i_col] += w
                            i_col += _lfsr_random_integers(state, 1, clen0 - 1)
        else:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, vector, seed, posts):
                posts[:] = 0.
                num_col = posts.shape[0]
                num_row = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                for i_row in range(num_row):
                    if vector[i_row] > 0.:
                        state = _lfsr_seed(seed0 + i_row * num_col)
                        i_col = _lfsr_random_integers(state, 0, clen0 - 1)
                        while i_col < num_col:
                            w = _lfsr_normal(state, w_loc0, w_scale0)
                            posts[i_col] += w
                            i_col += _lfsr_random_integers(state, 1, clen0 - 1)

    def run(w_loc, w_scale, clen, vector, seed):
        return numba_kernel(kernel, outs=kwargs['outs'])(w_loc, w_scale, clen, vector, seed)

    return run


def _jitc_mv_normal_warp_kernel_generator(
    w_loc_info: jax.ShapeDtypeStruct,
    w_scale_info: jax.ShapeDtypeStruct,
    clen_info: jax.ShapeDtypeStruct,
    vector_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    seed_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    r"""
    Generate the GPU kernel for the :func:`_jitc_matvec_normal` operation.

    JITC matrix generation must be consistent with _jitn_warp_kernel_generator in float.py:

    - corder=True:  i_row = tid(), loop over i_col, seed = seed0 + i_row
    - corder=False: i_col = tid(), loop over i_row, seed = seed0 + i_col
    """
    import warp
    from warp.jax_experimental import jax_kernel

    w_loc_warp_info = jaxinfo_to_warpinfo(w_loc_info)
    w_scale_warp_info = jaxinfo_to_warpinfo(w_scale_info)
    clen_warp_info = jaxinfo_to_warpinfo(clen_info)
    v_warp_info = jaxinfo_to_warpinfo(vector_info)
    seed_warp_info = jaxinfo_to_warpinfo(seed_info)
    out_warp_info = jaxinfo_to_warpinfo(out_info)

    if corder:
        # Consistent with jitn corder=True: i_row=tid(), loop over i_col
        # Each thread produces one output element posts[i_row] — no atomics needed.
        if vector_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                vector: v_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                num_col = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_row = warp.tid()
                r = float(0.0)
                state = warp.rand_init(seed0 + i_row * num_col)
                i_col = warp.randi(state, 0, clen0)
                while i_col < num_col:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    r = warp.where(vector[i_col], r + w, r)
                    i_col += warp.randi(state, 1, clen0)
                posts[i_row] = r

        else:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                vector: v_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                num_col = vector.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_row = warp.tid()
                r = float(0.0)
                state = warp.rand_init(seed0 + i_row * num_col)
                i_col = warp.randi(state, 0, clen0)
                while i_col < num_col:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    if vector[i_col] > float(0.0):
                        r += w
                    i_col += warp.randi(state, 1, clen0)
                posts[i_row] = r

        def run(w_loc, w_scale, clen, vector, seed):
            dim = out_info.shape[0]
            fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, output_dims={'posts': out_info.shape})
            return fn(w_loc, w_scale, clen, vector, seed)

    else:
        # Consistent with jitn corder=False: i_col=tid(), loop over i_row
        # Multiple threads may scatter into the same output row — must use atomic_add.
        if vector_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                vector: v_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                num_row = posts.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_col = warp.tid()
                v = vector[i_col]
                state = warp.rand_init(seed0 + i_col * num_row)
                i_row = warp.randi(state, 0, clen0)
                while i_row < num_row:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    if v:
                        warp.atomic_add(posts, i_row, w)
                    i_row += warp.randi(state, 1, clen0)

        else:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                vector: v_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                num_row = posts.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_col = warp.tid()
                v = vector[i_col]
                state = warp.rand_init(seed0 + i_col * num_row)
                i_row = warp.randi(state, 0, clen0)
                while i_row < num_row:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    if v > float(0.0):
                        warp.atomic_add(posts, i_row, w)
                    i_row += warp.randi(state, 1, clen0)

        def run(w_loc, w_scale, clen, vector, seed):
            dim = vector_info.shape[0]
            fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
            return fn(w_loc, w_scale, clen, vector, seed, jnp.zeros(out_info.shape, out_info.dtype))

    return run


def _jitc_mv_normal_pallas_kernel_generator(
    vector_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    """
    Pallas GPU kernel for binary event matvec with normal-distributed JITC matrix.

    JITC matrix generation must be consistent with _jitnmv_pallas_kernel_generator
    in float.py (same RNG seeding and iteration order):
    - corder=True:  vectorize over output rows (i_cols), loop over input (i_rows)
    - corder=False: vectorize over input rows (i_rows), loop over output (i_cols)
    """
    from jax.experimental import pallas as pl
    from jax.experimental.pallas.triton import atomic_add  # type: ignore[assignment]

    _PallasLFSRRNG = get_pallas_lfsr_rng_class()

    dim = (out_info.shape[0] if corder else vector_info.shape[0])
    block_size = generate_block_dim(dim, maximum=128)

    if corder:
        # Matches float.py _jitnmv_pallas corder=True exactly:
        # vectorize over output, seed by i_cols, loop over i_rows.
        # Binary: accumulate w only when vector[i_row] is event (>0 or True).
        def kernel(w_loc_ref, w_scale_ref, clen_ref, vector_ref, seed_ref, _, post_ref):
            num_row = vector_ref.shape[0]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen = clen_ref[0]
            seed = seed_ref[0]
            i_col_block = pl.program_id(0)
            i_cols = i_col_block * block_size + jnp.arange(block_size)
            i_col_mask = i_cols < dim

            def body(data):
                i_rows, i_row_mask, rng, out = data
                v = vector_ref[i_rows]
                if vector_ref.dtype != jnp.bool_:
                    v = v > 0.
                w = rng.normal(w_loc, w_scale)
                out = jnp.where(i_row_mask & v, out + w, out)
                i_rows += rng.random_integers(1, clen)
                return i_rows, i_rows < num_row, rng, out

            rng = _PallasLFSRRNG(seed + i_cols * num_row)
            i_rows = rng.random_integers(0, clen)
            i_row_mask = i_rows < num_row
            out = jnp.zeros(block_size, dtype=post_ref.dtype)
            out = jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_rows, i_row_mask, rng, out)
            )[-1]
            post_ref[i_cols] = jnp.where(i_col_mask, out, post_ref[i_cols])

    else:
        # Matches float.py _jitnmv_pallas corder=False exactly:
        # vectorize over input, seed by i_rows, loop over i_cols.
        # Binary: only scatter w (via atomic_add) when vector element is event.
        def kernel(w_loc_ref, w_scale_ref, clen_ref, vector_ref, seed_ref, _, post_ref):
            num_col = post_ref.shape[0]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen = clen_ref[0]
            seed = seed_ref[0]
            i_row_block = pl.program_id(0)
            i_rows = i_row_block * block_size + jnp.arange(block_size)
            i_row_mask = i_rows < dim
            v = vector_ref[i_rows]
            if vector_ref.dtype != jnp.bool_:
                v = v > 0.
            # event_mask: only active lanes where the vector element is an event
            event_mask = i_row_mask & v

            def body(data):
                i_cols, i_col_mask, rng = data
                w = rng.normal(w_loc, w_scale)
                atomic_add(post_ref, (i_cols,), w, mask=event_mask & i_col_mask)
                i_cols += rng.random_integers(1, clen)
                return i_cols, i_cols < num_col, rng

            rng = _PallasLFSRRNG(seed + i_rows * num_col)
            i_cols = rng.random_integers(0, clen)
            i_col_mask = i_cols < num_col
            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_cols, i_col_mask, rng)
            )

    def run(w_loc, w_scale, clen, vector, seed):
        fn = pl.pallas_call(
            kernel,
            grid=(pl.cdiv(dim, block_size),),
            input_output_aliases={5: 0},
            out_shape=kwargs['outs'],
            backend='triton',
        )
        placeholder = jnp.zeros(kwargs['outs'][0].shape, kwargs['outs'][0].dtype)
        return fn(w_loc, w_scale, clen, vector, seed, placeholder)

    return run


def _jitc_mv_normal_jvp_v(v_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the vector argument of binary_jitnmv."""
    return jitnmv_p_call(
        w_loc, w_scale, clen, v_dot, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mv_normal_jvp_wloc(w_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the w_loc argument of binary_jitnmv."""
    return binary_jitnmv_p_call(
        w_dot, w_scale, clen, vector, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mv_normal_jvp_wscale(w_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the w_scale argument of binary_jitnmv."""
    return binary_jitnmv_p_call(
        w_loc, w_dot, clen, vector, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mv_normal_transpose_rules(ct, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, **kwargs):
    """Transpose (VJP) rule for binary_jitnmv."""
    assert not ad.is_undefined_primal(clen)
    assert not ad.is_undefined_primal(seed)

    ct = ct[0]
    if ad.is_undefined_primal(vector):
        r = jitnmv_p_call(
            w_loc,
            w_scale,
            clen,
            ct,
            seed,
            shape=shape,
            transpose=not transpose,
            corder=not corder,
            backend=kwargs['backend'],
        )[0]
        return w_loc, w_scale, clen, r, seed
    elif ad.is_undefined_primal(w_loc):
        # M = (w_loc + w_scale * Z) * mask, forward: M @ event(v)
        # d(loss)/d(w_loc) = sum((mask^T @ ct) * vector)
        r = jitnmv_p_call(
            1., 0., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=kwargs['backend'],
        )[0]
        dw_loc = jnp.expand_dims(jnp.sum(r * vector), axis=0)
        return dw_loc, w_scale, clen, vector, seed
    elif ad.is_undefined_primal(w_scale):
        # d(loss)/d(w_scale) = sum(((Z*mask)^T @ ct) * vector)
        r = jitnmv_p_call(
            0., 1., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=kwargs['backend'],
        )[0]
        dw_scale = jnp.expand_dims(jnp.sum(r * vector), axis=0)
        return w_loc, dw_scale, clen, vector, seed
    else:
        raise NotImplementedError(
            f"Transpose rule for binary_jitnmv not implemented "
            f"when none of vector/w_loc/w_scale is an undefined primal."
        )


def _jitc_mv_normal_batching(args, axes, **kwargs):
    """Batching rule for binary_jitnmv, promoting to binary_jitnmm when batched."""
    if tuple(axes) == (None, None, None, 0, None):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = binary_jitnmm_p_call(
            args[0],
            args[1],
            args[2],
            args[3].T,
            args[4],
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            corder=kwargs['corder'],
            backend=kwargs['backend'],
        )
        return r, [1]
    elif tuple(axes) == (None, None, None, 1, None):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = binary_jitnmm_p_call(
            args[0],
            args[1],
            args[2],
            args[3],
            args[4],
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            corder=kwargs['corder'],
            backend=kwargs['backend'],
        )
        return r, [1]
    else:
        return general_batching_rule(
            binary_jitnmv_p,
            args,
            axes,
            **kwargs,
        )


def _binary_jitnmv_benchmark_data(*, platform):
    """Generate benchmark configurations for binary_jitnmv."""
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for corder in (True, False):
            for bool_event in (True, False):
                w_loc = jnp.ones(1, dtype=dtype)
                w_scale = jnp.ones(1, dtype=dtype) * 0.1
                clen = jnp.atleast_1d(jnp.asarray(2.0 / prob, dtype=dtype))
                v_size = n_post if not transpose else n_pre
                if bool_event:
                    vector = jnp.asarray(np.random.rand(v_size) > 0.5, dtype=jnp.bool_)
                else:
                    vector = jnp.asarray(np.random.rand(v_size), dtype=dtype)
                seed = jnp.asarray(42, dtype=jnp.uint32)
                name = f"{'T' if transpose else 'NT'},{'corder' if corder else 'rorder'},{'bool' if bool_event else 'float'}"
                configs.append(BenchmarkConfig(name, (w_loc, w_scale, clen, vector, seed), {
                    'shape': (n_pre, n_post), 'transpose': transpose, 'corder': corder
                }))
    return configs


def binary_jitnmv_p_call(
    w_loc,
    w_scale,
    clen,
    vector,
    seed,
    *,
    shape: Sequence[int],
    transpose: bool,
    corder: bool,
    backend: Optional[str] = None,
):
    """
    Low-level primitive call for event-driven matrix-vector multiplication with normal weights.

    This function validates inputs, computes the output shape, and dispatches to
    the ``binary_jitnmv_p`` XLA custom kernel. It is called by ``binary_jitnmv``
    after unit handling and seed initialization, and is also used directly by
    JVP, transpose, and batching rules.

    Parameters
    ----------
    w_loc : jax.Array
        Location (mean) parameter, shape ``(1,)``.
    w_scale : jax.Array
        Scale (standard deviation) parameter, shape ``(1,)``.
    clen : jax.Array
        Connection length (approximately ``2 / prob``), shape ``(1,)``.
    vector : jax.Array
        1-D binary event vector.
    seed : jax.Array
        Random seed, shape ``(1,)``.
    shape : Sequence[int]
        Shape of the implicit connectivity matrix ``(m, n)``.
    transpose : bool
        If True, compute ``M.T @ v`` instead of ``M @ v``.
    corder : bool
        Memory layout order flag for kernel dispatch.
    backend : str, optional
        Compute backend override.

    Returns
    -------
    tuple
        A single-element tuple containing the output vector as a JAX array.

    Raises
    ------
    AssertionError
        If input shapes or dimensions are incompatible.

    See Also
    --------
    binary_jitnmv : High-level API with unit handling.
    binary_jitnmv_p : The underlying XLA custom kernel primitive.
    """
    w_loc = jnp.atleast_1d(w_loc)
    w_scale = jnp.atleast_1d(w_scale)
    clen = jnp.atleast_1d(clen)

    assert len(shape) == 2, "The matrix shape should be a tuple of two integers."
    assert w_loc.shape == (1,), f"The weight shape should be (1,), but got {w_loc.shape}."
    assert w_scale.shape == (1,), f"The weight shape should be (1,), but got {w_scale.shape}."
    assert clen.shape == (1,), f"The clen shape should be (1,), but got {clen.shape}."
    assert vector.ndim == 1, f"The vector should be a 1D array, but got {vector.ndim}D."
    assert seed.shape == (1,), f"The seed shape should be (1,), but got {seed.shape}."

    if transpose:
        assert shape[0] == len(vector), f"The matrix shape and vector length do not match. {vector.shape} @ {shape}"
    else:
        assert shape[1] == len(vector), f"The matrix shape and vector length do not match. {shape} @ {vector.shape}"

    out_info = (
        jax.ShapeDtypeStruct([shape[1]], w_loc.dtype)
        if transpose else
        jax.ShapeDtypeStruct([shape[0]], w_loc.dtype)
    )

    return binary_jitnmv_p(
        w_loc,
        w_scale,
        clen,
        vector,
        seed,
        outs=[out_info],
        w_loc_info=jax.ShapeDtypeStruct(w_loc.shape, w_loc.dtype),
        w_scale_info=jax.ShapeDtypeStruct(w_scale.shape, w_scale.dtype),
        clen_info=jax.ShapeDtypeStruct(clen.shape, clen.dtype),
        vector_info=jax.ShapeDtypeStruct(vector.shape, vector.dtype),
        seed_info=jax.ShapeDtypeStruct(seed.shape, seed.dtype),
        out_info=out_info,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )


binary_jitnmv_p = XLACustomKernel(
    'event_jitc_mv_normal',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_jitnmv``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) JIT normal connectivity
matrix-vector multiplication operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

In this operation, the connectivity matrix has weights normally distributed with specified
mean and standard deviation, and the input vector is treated as binary events (spikes).
Only active events contribute to the output computation.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_jitnmv_p.available_backends(platform)``,
and the default backend can be configured with ``binary_jitnmv_p.set_default(platform, backend)``.

See Also
--------
binary_jitnmv : High-level user-facing function wrapper.
"""
)
binary_jitnmv_p.def_numba_kernel(_jitc_mv_normal_numba_kernel_generator)
binary_jitnmv_p.def_warp_kernel(_jitc_mv_normal_warp_kernel_generator)
binary_jitnmv_p.def_pallas_kernel('gpu', _jitc_mv_normal_pallas_kernel_generator)
binary_jitnmv_p.def_jvp_rule2(_jitc_mv_normal_jvp_wloc, _jitc_mv_normal_jvp_wscale, None, _jitc_mv_normal_jvp_v, None)
binary_jitnmv_p.def_transpose_rule(_jitc_mv_normal_transpose_rules)
binary_jitnmv_p.def_batching_rule(_jitc_mv_normal_batching)
binary_jitnmv_p.def_tags('jit_normal', 'binary')
binary_jitnmv_p.def_benchmark_data(_binary_jitnmv_benchmark_data)


def _jitc_mm_normal_numba_kernel_generator(
    transpose: bool,
    corder: bool,
    B_info: jax.ShapeDtypeStruct,
    **kwargs
):
    r"""
    Generate the CPU kernel for the :func:`_jitc_matmat_normal` operation.
    """
    import numba
    _lfsr_seed = get_numba_lfsr_seed()
    _lfsr_random_integers = get_numba_lfsr_random_integers()
    _lfsr_normal = get_numba_lfsr_normal()

    if corder:
        # JIT Matrix.T @ B
        # - JIT matrix: [k, m]
        # - B: [k, n]
        if B_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, B, seed, posts):
                posts[:] = 0.
                m = posts.shape[0]
                n = posts.shape[1]
                k = B.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                seed0 = seed[0]
                clen0 = clen[0]
                for i_m in range(m):
                    state = _lfsr_seed(seed0 + i_m * k)
                    i_k = _lfsr_random_integers(state, 0, clen0 - 1)
                    out = np.zeros(n, dtype=posts.dtype)
                    while i_k < k:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        for j in range(B.shape[1]):
                            if B[i_k, j]:
                                out[j] += w
                        i_k += _lfsr_random_integers(state, 1, clen0 - 1)
                    posts[i_m] = out
        else:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, B, seed, posts):
                posts[:] = 0.
                m = posts.shape[0]
                n = posts.shape[1]
                k = B.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                seed0 = seed[0]
                clen0 = clen[0]
                for i_m in range(m):
                    state = _lfsr_seed(seed0 + i_m * k)
                    i_k = _lfsr_random_integers(state, 0, clen0 - 1)
                    out = np.zeros(n, dtype=posts.dtype)
                    while i_k < k:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        for j in range(B.shape[1]):
                            if B[i_k, j] > 0.:
                                out[j] += w
                        i_k += _lfsr_random_integers(state, 1, clen0 - 1)
                    posts[i_m] = out

    else:
        # JIT Matrix.T @ B
        # - JIT matrix: [k, m]
        # - B: [k, n]
        if B_info.dtype == jnp.bool_:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, B, seed, posts):
                posts[:] = 0.
                m = posts.shape[0]
                k = B.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                seed0 = seed[0]
                clen0 = clen[0]
                for i_k in range(k):
                    state = _lfsr_seed(seed0 + i_k * m)
                    indices = np.where(B[i_k])[0]
                    i_m = _lfsr_random_integers(state, 0, clen0 - 1)
                    while i_m < m:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        posts[i_m, indices] += w
                        i_m += _lfsr_random_integers(state, 1, clen0 - 1)
        else:
            @numba.njit(fastmath=True)
            def kernel(w_loc, w_scale, clen, B, seed, posts):
                posts[:] = 0.
                m = posts.shape[0]
                k = B.shape[0]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                seed0 = seed[0]
                clen0 = clen[0]
                for i_k in range(k):
                    state = _lfsr_seed(seed0 + i_k * m)
                    indices = np.where(B[i_k] > 0.)[0]
                    i_m = _lfsr_random_integers(state, 0, clen0 - 1)
                    while i_m < m:
                        w = _lfsr_normal(state, w_loc0, w_scale0)
                        posts[i_m, indices] += w
                        i_m += _lfsr_random_integers(state, 1, clen0 - 1)

    def run(w_loc, w_scale, clen, B, seed):
        return numba_kernel(kernel, outs=kwargs['outs'])(w_loc, w_scale, clen, B, seed)

    return run


def _jitc_mm_normal_warp_kernel_generator(
    w_loc_info: jax.ShapeDtypeStruct,
    w_scale_info: jax.ShapeDtypeStruct,
    clen_info: jax.ShapeDtypeStruct,
    B_info: jax.ShapeDtypeStruct,
    seed_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    r"""
    Generate the GPU kernel for the :func:`_jitc_matmat_normal` operation.

    Uses scalar loops instead of warp tile operations to avoid cooperative
    warp issues. JITC matrix generation must be consistent with
    _jitn_warp_kernel_generator in float.py:
    - corder=True:  i_m = tid(), loop over i_k, seed = seed0 + i_m
    - corder=False: i_k = tid(), loop over i_m, seed = seed0 + i_k
    """
    import warp
    from warp.jax_experimental import jax_kernel

    w_loc_warp_info = jaxinfo_to_warpinfo(w_loc_info)
    w_scale_warp_info = jaxinfo_to_warpinfo(w_scale_info)
    clen_warp_info = jaxinfo_to_warpinfo(clen_info)
    B_warp_info = jaxinfo_to_warpinfo(B_info)
    seed_warp_info = jaxinfo_to_warpinfo(seed_info)
    out_warp_info = jaxinfo_to_warpinfo(out_info)

    if corder:
        # Consistent with jitn corder=True: i_m=tid(), loop over i_k
        # Each thread produces one output row — no atomics needed.
        if B_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                B: B_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                k = B.shape[0]
                n = B.shape[1]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_m = warp.tid()
                state = warp.rand_init(seed0 + i_m * k)
                i_k = warp.randi(state, 0, clen0)
                while i_k < k:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    for j in range(n):
                        if B[i_k, j]:
                            posts[i_m, j] += w
                    i_k += warp.randi(state, 1, clen0)

        else:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                B: B_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                k = B.shape[0]
                n = B.shape[1]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_m = warp.tid()
                state = warp.rand_init(seed0 + i_m * k)
                i_k = warp.randi(state, 0, clen0)
                while i_k < k:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    for j in range(n):
                        if B[i_k, j] > float(0.0):
                            posts[i_m, j] += w
                    i_k += warp.randi(state, 1, clen0)

        def run(w_loc, w_scale, clen, B, seed):
            dim = out_info.shape[0]
            fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
            return fn(w_loc, w_scale, clen, B, seed, jnp.zeros(out_info.shape, out_info.dtype))

    else:
        # Consistent with jitn corder=False: i_k=tid(), loop over i_m
        # Multiple threads scatter into output rows — must use atomic_add.
        if B_info.dtype == jnp.bool_:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                B: B_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                m = posts.shape[0]
                n = B.shape[1]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_k = warp.tid()
                state = warp.rand_init(seed0 + i_k * m)
                i_m = warp.randi(state, 0, clen0)
                while i_m < m:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    for j in range(n):
                        if B[i_k, j]:
                            warp.atomic_add(posts, i_m, j, w)
                    i_m += warp.randi(state, 1, clen0)

        else:
            @warp.kernel
            def kernel(
                w_loc: w_loc_warp_info,
                w_scale: w_scale_warp_info,
                clen: clen_warp_info,
                B: B_warp_info,
                seed: seed_warp_info,
                posts: out_warp_info,
            ):
                m = posts.shape[0]
                n = B.shape[1]
                w_loc0 = w_loc[0]
                w_scale0 = w_scale[0]
                clen0 = clen[0]
                seed0 = seed[0]
                i_k = warp.tid()
                state = warp.rand_init(seed0 + i_k * m)
                i_m = warp.randi(state, 0, clen0)
                while i_m < m:
                    w = warp.randn(state) * w_scale0 + w_loc0
                    for j in range(n):
                        if B[i_k, j] > float(0.0):
                            warp.atomic_add(posts, i_m, j, w)
                    i_m += warp.randi(state, 1, clen0)

        def run(w_loc, w_scale, clen, B, seed):
            dim = B_info.shape[0]
            fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
            return fn(w_loc, w_scale, clen, B, seed, jnp.zeros(out_info.shape, out_info.dtype))

    return run


def _jitc_mm_normal_pallas_kernel_generator(
    B_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    """
    Pallas GPU kernel for binary event matmat with normal-distributed JITC matrix.

    Matches _jitnmm_pallas_kernel_generator in float.py:
    - Grid: (row_or_k_blocks, B_cols) — each block processes one B column
    - corder=True:  vectorize over output rows, seed by i_rows, loop over k
    - corder=False: vectorize over k, seed by i_ks, loop over output rows
    """
    from jax.experimental import pallas as pl
    from jax.experimental.pallas.triton import atomic_add  # type: ignore[assignment]

    _PallasLFSRRNG = get_pallas_lfsr_rng_class()

    B_cols = B_info.shape[1]

    if corder:
        # Match float.py _jitnmm_pallas corder=True exactly:
        # Grid: (row_blocks, B_cols). Each block processes one B column.
        out_rows = out_info.shape[0]
        row_block = generate_block_dim(out_rows, maximum=128)
        grid = (pl.cdiv(out_rows, row_block), B_cols)

        def kernel(w_loc_ref, w_scale_ref, clen_ref, B_ref, seed_ref, _, post_ref):
            k = B_ref.shape[0]
            w_loc0 = w_loc_ref[0]
            w_scale0 = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_row_block = pl.program_id(0)
            col_j = pl.program_id(1)

            i_rows = i_row_block * row_block + jnp.arange(row_block)
            i_row_mask = i_rows < out_rows
            safe_rows = jnp.where(i_row_mask, i_rows, 0)

            rng = _PallasLFSRRNG(seed0 + i_rows * k)
            i_cols = rng.random_integers(0, clen0)
            i_col_mask = i_cols < k

            out = jnp.zeros(row_block, dtype=post_ref.dtype)

            def body(data):
                i_cols, i_col_mask, rng, out = data
                w = rng.normal(w_loc0, w_scale0)
                safe_cols = jnp.where(i_col_mask, i_cols, 0)
                b_vals = B_ref[safe_cols, col_j]
                # Binary thresholding: treat b_vals as events
                if B_ref.dtype == jnp.bool_:
                    events = jnp.asarray(b_vals, dtype=out.dtype)
                else:
                    events = jnp.where(b_vals > 0., 1., 0.)
                out += jnp.where(i_col_mask & i_row_mask, w * events, 0.)
                i_cols += rng.random_integers(1, clen0)
                return i_cols, i_cols < k, rng, out

            _, _, _, out = jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_cols, i_col_mask, rng, out)
            )
            atomic_add(post_ref, (safe_rows, col_j), out, mask=i_row_mask)

    else:
        # Match float.py _jitnmm_pallas corder=False exactly:
        # Grid: (k_blocks, B_cols). Each block processes one B column.
        k_dim = B_info.shape[0]
        k_block = generate_block_dim(k_dim, maximum=128)
        grid = (pl.cdiv(k_dim, k_block), B_cols)

        def kernel(w_loc_ref, w_scale_ref, clen_ref, B_ref, seed_ref, _, post_ref):
            m = post_ref.shape[0]
            w_loc0 = w_loc_ref[0]
            w_scale0 = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_k_block = pl.program_id(0)
            col_j = pl.program_id(1)

            i_ks = i_k_block * k_block + jnp.arange(k_block)
            i_k_mask = i_ks < k_dim
            safe_ks = jnp.where(i_k_mask, i_ks, 0)

            # Preload B values for this column and apply binary thresholding
            b_vals = B_ref[safe_ks, col_j]
            if B_ref.dtype == jnp.bool_:
                b_events = jnp.asarray(b_vals, dtype=post_ref.dtype)
            else:
                b_events = jnp.where(b_vals > 0., 1., 0.)
            b_events = jnp.where(i_k_mask, b_events, 0.)

            rng = _PallasLFSRRNG(seed0 + i_ks * m)
            i_rows = rng.random_integers(0, clen0)
            i_row_mask = i_rows < m

            def body(data):
                i_rows, i_row_mask, rng = data
                w = rng.normal(w_loc0, w_scale0)
                vals = jnp.where(i_k_mask & i_row_mask, w * b_events, 0.)
                safe_rows = jnp.where(i_row_mask, i_rows, 0)
                atomic_add(post_ref, (safe_rows, col_j), vals,
                           mask=i_k_mask & i_row_mask)
                i_rows += rng.random_integers(1, clen0)
                return i_rows, i_rows < m, rng

            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_rows, i_row_mask, rng)
            )

    def run(w_loc, w_scale, clen, B, seed):
        fn = pl.pallas_call(
            kernel,
            grid=grid,
            input_output_aliases={5: 0},
            out_shape=kwargs['outs'],
            backend='triton',
        )
        placeholder = jnp.zeros(kwargs['outs'][0].shape, kwargs['outs'][0].dtype)
        return fn(w_loc, w_scale, clen, B, seed, placeholder)

    return run


def _jitc_mm_normal_jvp_wloc(w_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the w_loc argument of binary_jitnmm."""
    return binary_jitnmm_p_call(
        w_dot, w_scale, clen, B, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mm_normal_jvp_wscale(w_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the w_scale argument of binary_jitnmm."""
    return binary_jitnmm_p_call(
        w_loc, w_dot, clen, B, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mm_normal_jvp_B(B_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, **kwargs):
    """JVP rule for the B argument of binary_jitnmm."""
    return jitnmm_p_call(
        w_loc, w_scale, clen, B_dot, seed, shape=shape, transpose=transpose, corder=corder, backend=kwargs['backend'],
    )


def _jitc_mm_normal_transpose_rules(ct, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, **kwargs):
    """Transpose (VJP) rule for binary_jitnmm."""
    assert not ad.is_undefined_primal(clen)
    assert not ad.is_undefined_primal(seed)

    ct = ct[0]
    if ad.is_undefined_primal(B):
        r = jitnmm_p_call(
            w_loc,
            w_scale,
            clen,
            ct,
            seed,
            shape=shape,
            transpose=not transpose,
            corder=not corder,
            backend=kwargs['backend'],
        )[0]
        return w_loc, w_scale, clen, r, seed
    elif ad.is_undefined_primal(w_loc):
        # M = (w_loc + w_scale * Z) * mask, forward: M @ event(B)
        # d(loss)/d(w_loc) = sum((mask^T @ ct) * B)
        r = jitnmm_p_call(
            1., 0., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=kwargs['backend'],
        )[0]
        dw_loc = jnp.expand_dims(jnp.sum(r * B), axis=0)
        return dw_loc, w_scale, clen, B, seed
    elif ad.is_undefined_primal(w_scale):
        # d(loss)/d(w_scale) = sum(((Z*mask)^T @ ct) * B)
        r = jitnmm_p_call(
            0., 1., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=kwargs['backend'],
        )[0]
        dw_scale = jnp.expand_dims(jnp.sum(r * B), axis=0)
        return w_loc, dw_scale, clen, B, seed
    else:
        raise NotImplementedError(
            'Transpose rules for binary_jitc_matmat_normal not implemented for '
            'non-undefined primals.'
        )


def _batching_axis1(args, axis=1, **kwargs):
    """Helper for batching binary_jitnmm along axis 1 (or transposed axis 0)."""
    assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
    m, maybe_batch1, maybe_batch2 = args[3].shape
    B = args[3].reshape(m, maybe_batch1 * maybe_batch2)
    r = binary_jitnmm_p_call(
        args[0],
        args[1],
        args[2],
        B,
        args[4],
        shape=kwargs['shape'],
        transpose=kwargs['transpose'],
        corder=kwargs['corder'],
        backend=kwargs['backend'],
    )
    r = jnp.reshape(r[0], [r[0].shape[0], maybe_batch1, maybe_batch2])
    return [r], [axis]


def _jitc_mm_normal_batching(args, axes, **kwargs):
    """Batching rule for binary_jitnmm."""
    if tuple(axes) == (None, None, None, 0, None):
        assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
        args = list(args)
        args[3] = jnp.transpose(args[3], (1, 0, 2))
        return _batching_axis1(args, **kwargs)

    elif tuple(axes) == (None, None, None, 1, None):
        return _batching_axis1(args, **kwargs)

    elif tuple(axes) == (None, None, None, 2, None):
        return _batching_axis1(args, axis=2, **kwargs)

    else:
        return general_batching_rule(binary_jitnmm_p, args, axes, **kwargs)


def _binary_jitnmm_benchmark_data(*, platform):
    """Generate benchmark configurations for binary_jitnmm."""
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for corder in (True, False):
            for bool_event in (True, False):
                w_loc = jnp.ones(1, dtype=dtype)
                w_scale = jnp.ones(1, dtype=dtype) * 0.1
                clen = jnp.atleast_1d(jnp.asarray(2.0 / prob, dtype=dtype))
                b_rows = n_post if not transpose else n_pre
                if bool_event:
                    B = jnp.asarray(np.random.rand(b_rows, 10) > 0.5, dtype=jnp.bool_)
                else:
                    B = jnp.asarray(np.random.rand(b_rows, 10), dtype=dtype)
                seed = jnp.asarray(42, dtype=jnp.uint32)
                name = f"{'T' if transpose else 'NT'},{'corder' if corder else 'rorder'},{'bool' if bool_event else 'float'}"
                configs.append(
                    BenchmarkConfig(
                        name,
                        (w_loc, w_scale, clen, B, seed),
                        {'shape': (n_pre, n_post), 'transpose': transpose, 'corder': corder}
                    )
                )
    return configs


def binary_jitnmm_p_call(
    w_loc,
    w_scale,
    clen,
    B,
    seed,
    *,
    shape: MatrixShape,
    transpose: bool,
    corder: bool,
    backend: Optional[str] = None,
):
    """
    Low-level primitive call for event-driven matrix-matrix multiplication with normal weights.

    This function validates inputs, computes the output shape, and dispatches to
    the ``binary_jitnmm_p`` XLA custom kernel. It is called by ``binary_jitnmm``
    after unit handling and seed initialization, and is also used directly by
    JVP, transpose, and batching rules.

    Parameters
    ----------
    w_loc : jax.Array
        Location (mean) parameter, shape ``(1,)``.
    w_scale : jax.Array
        Scale (standard deviation) parameter, shape ``(1,)``.
    clen : jax.Array
        Connection length (approximately ``2 / prob``), shape ``(1,)``.
    B : jax.Array
        2-D binary event matrix, shape ``(k, n)``.
    seed : jax.Array
        Random seed, shape ``(1,)``.
    shape : MatrixShape
        Shape of the implicit connectivity matrix ``(m, k)`` (or ``(k, m)`` when
        ``transpose=True``).
    transpose : bool
        If True, compute ``M.T @ B`` instead of ``M @ B``.
    corder : bool
        Memory layout order flag for kernel dispatch.
    backend : str, optional
        Compute backend override.

    Returns
    -------
    tuple
        A single-element tuple containing the output matrix as a JAX array.

    Raises
    ------
    AssertionError
        If input shapes or dimensions are incompatible.

    See Also
    --------
    binary_jitnmm : High-level API with unit handling.
    binary_jitnmm_p : The underlying XLA custom kernel primitive.
    """
    w_loc = jnp.atleast_1d(w_loc)
    w_scale = jnp.atleast_1d(w_scale)
    clen = jnp.atleast_1d(clen)

    assert len(shape) == 2, "The matrix shape should be a tuple of two integers."
    assert B.ndim == 2, "The input matrix B should be a 2D array."
    assert seed.ndim == 1, "The seed should be a 1D array."
    assert w_loc.ndim == 1, "The weight should be a 1D array."
    assert w_scale.ndim == 1, "The weight should be a 1D array."
    assert clen.ndim == 1, "The clen should be a 1D array."
    assert w_loc.shape == (1,), "The weight should be a scalar."
    assert w_scale.shape == (1,), "The weight should be a scalar."
    assert clen.shape == (1,), "The clen should be a scalar."
    assert seed.shape == (1,), "The seed should be a scalar."
    if transpose:
        assert shape[0] == B.shape[0], f"The matrix shape and B shape do not match. {B.shape} @ {shape}"
    else:
        assert shape[1] == B.shape[0], f"The matrix shape and B shape do not match. {shape} @ {B.shape}"

    out_info = (
        jax.ShapeDtypeStruct([shape[1], B.shape[1]], w_loc.dtype)
        if transpose else
        jax.ShapeDtypeStruct([shape[0], B.shape[1]], w_loc.dtype)
    )

    return binary_jitnmm_p(
        w_loc,
        w_scale,
        clen,
        B,
        seed,
        outs=[out_info],
        w_loc_info=jax.ShapeDtypeStruct(w_loc.shape, w_loc.dtype),
        w_scale_info=jax.ShapeDtypeStruct(w_scale.shape, w_scale.dtype),
        clen_info=jax.ShapeDtypeStruct(clen.shape, clen.dtype),
        B_info=jax.ShapeDtypeStruct(B.shape, B.dtype),
        seed_info=jax.ShapeDtypeStruct(seed.shape, seed.dtype),
        out_info=out_info,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )


binary_jitnmm_p = XLACustomKernel(
    'binary_jitc_mm_normal',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_jitnmm``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) JIT normal connectivity
matrix-matrix multiplication operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

In this operation, the connectivity matrix has weights normally distributed with specified
mean and standard deviation, and the input matrix is treated as binary events (spikes).
Each column of the input is processed independently as an event vector.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_jitnmm_p.available_backends(platform)``,
and the default backend can be configured with ``binary_jitnmm_p.set_default(platform, backend)``.

See Also
--------
binary_jitnmm : High-level user-facing function wrapper.
"""
)
binary_jitnmm_p.def_numba_kernel(_jitc_mm_normal_numba_kernel_generator)
binary_jitnmm_p.def_warp_kernel(_jitc_mm_normal_warp_kernel_generator)
binary_jitnmm_p.def_pallas_kernel('gpu', _jitc_mm_normal_pallas_kernel_generator)
binary_jitnmm_p.def_jvp_rule2(_jitc_mm_normal_jvp_wloc, _jitc_mm_normal_jvp_wscale, None, _jitc_mm_normal_jvp_B, None)
binary_jitnmm_p.def_transpose_rule(_jitc_mm_normal_transpose_rules)
binary_jitnmm_p.def_batching_rule(_jitc_mm_normal_batching)
binary_jitnmm_p.def_tags('jit_normal', 'binary')
binary_jitnmm_p.def_benchmark_data(_binary_jitnmm_benchmark_data)



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""
This module defines several type aliases using the `typing` module.

Type Aliases:
    MatrixShape: A tuple representing the shape of a matrix, with two integers.
    Data: A union type that can be a `jax.Array`, `numpy.ndarray`, or `brainunit.Quantity`.
    Index: A union type that can be either a `jax.Array` or `numpy.ndarray`.
    Row: Alias for Index, representing a row index.
    Col: Alias for Index, representing a column index.
    Indptr: Alias for Index, representing an index pointer.
    Kernel: A callable type, representing a function or method.
"""

import numbers
from typing import Union, Tuple, Callable

import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np

# A tuple representing the shape of a matrix, with two integers.
MatrixShape = Tuple[int, int]

# A union type that can be a jax.Array, numpy.ndarray, or brainunit.Quantity.
Data = Union[jax.Array, np.ndarray, u.Quantity, numbers.Number]

# A union type that can be either a jax.Array or numpy.ndarray.
Index = Union[jax.Array, np.ndarray]

# Alias for Index, representing a row index.
Row = Index

# Alias for Index, representing a column index.
Col = Index

# Alias for Index, representing an index pointer.
Indptr = Index

# A callable type, representing a function or method.
Kernel = Callable

# kernel function generator
KernelGenerator = Callable[..., Kernel]

# Represents a scalar weight value, which can be a number, NumPy array, JAX array, or BrainUnit quantity.
WeightScalar = Union[numbers.Number, np.ndarray, jax.Array, u.Quantity]

# Represents a probability value, which can be a float, NumPy array, or JAX array.
Prob = Union[float, np.ndarray, jax.Array]

# Represents a seed value for random number generation, which can be an integer, NumPy array, or JAX array.
Seed = Union[int, np.ndarray, jax.Array]

# Represents a random key for Pallas, which is a tuple of four uint32 values.
PallasRandomKey = Tuple[jnp.uint32, jnp.uint32, jnp.uint32, jnp.uint32]  # uint32, 4



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import os

import pytest

from brainevent._cli import _build_parser, _filter_primitives, main


class TestBuildParser:
    def test_parser_creates(self):
        parser = _build_parser()
        assert parser is not None

    def test_benchmark_performance_args(self):
        parser = _build_parser()
        args = parser.parse_args([
            'benchmark-performance',
            '--platform', 'cpu',
            '--data', 'csr',
        ])
        assert args.command == 'benchmark-performance'
        assert args.platform == 'cpu'
        assert args.data == 'csr'

    def test_defaults(self):
        parser = _build_parser()
        args = parser.parse_args([
            'benchmark-performance',
            '--platform', 'gpu',
        ])
        assert args.data == 'all'
        assert args.n_warmup == 5
        assert args.n_runs == 20
        assert args.output is None


class TestFilterPrimitives:
    def test_all_returns_everything(self):
        registry = {'a': object(), 'b': object()}
        result = _filter_primitives(registry, 'all')
        assert result == registry

    def test_filter_by_tag(self):
        class FakePrim:
            def __init__(self, tags):
                self._tags = tags

        registry = {
            'csr_a': FakePrim({'csr', 'binary'}),
            'coo_b': FakePrim({'coo', 'binary'}),
            'dense_c': FakePrim({'dense', 'float'}),
        }
        result = _filter_primitives(registry, 'csr')
        assert 'csr_a' in result
        assert 'coo_b' not in result
        assert 'dense_c' not in result

    def test_filter_comma_separated(self):
        class FakePrim:
            def __init__(self, tags):
                self._tags = tags

        registry = {
            'csr_a': FakePrim({'csr', 'binary'}),
            'coo_b': FakePrim({'coo', 'binary'}),
        }
        result = _filter_primitives(registry, 'csr,coo')
        assert 'csr_a' in result
        assert 'coo_b' in result

    def test_filter_no_match(self):
        class FakePrim:
            def __init__(self, tags):
                self._tags = tags

        registry = {
            'csr_a': FakePrim({'csr', 'binary'}),
        }
        result = _filter_primitives(registry, 'nonexistent')
        assert len(result) == 0


class TestMainEntryPoint:
    def test_no_command_returns_0(self):
        assert main([]) == 0

    def test_invalid_command(self):
        # argparse will error on invalid subcommand
        with pytest.raises(SystemExit):
            main(['invalid-command'])

    def test_missing_required_platform(self):
        with pytest.raises(SystemExit):
            main(['benchmark-performance'])


class TestXLACustomKernelPersist:
    def test_set_default_persist(self, tmp_path, monkeypatch):
        """set_default with persist=True should write to config."""
        config_path = str(tmp_path / 'brainevent' / 'defaults.json')
        monkeypatch.setattr('brainevent.config.get_config_path', lambda: config_path)
        from brainevent.config import invalidate_cache
        invalidate_cache()

        import brainevent
        prim = brainevent.binary_csrmv_p
        prim.set_default('cpu', 'numba', persist=True)

        assert os.path.isfile(config_path)
        with open(config_path) as f:
            data = json.load(f)
        assert data['defaults']['binary_csrmv']['cpu'] == 'numba'
        invalidate_cache()

    def test_set_default_no_persist(self, tmp_path, monkeypatch):
        """set_default without persist should not write to config."""
        config_path = str(tmp_path / 'brainevent' / 'defaults.json')
        monkeypatch.setattr('brainevent.config.get_config_path', lambda: config_path)
        from brainevent.config import invalidate_cache
        invalidate_cache()

        import brainevent
        prim = brainevent.binary_csrmv_p
        prim.set_default('cpu', 'numba', persist=False)

        assert not os.path.isfile(config_path)
        invalidate_cache()


class TestXLACustomKernelTags:
    def test_def_tags(self):
        import brainevent
        prim = brainevent.binary_csrmv_p
        assert 'csr' in prim._tags
        assert 'binary' in prim._tags

    def test_def_benchmark_data(self):
        import brainevent
        prim = brainevent.binary_csrmv_p
        assert prim._benchmark_data_fn is not None
        assert callable(prim._benchmark_data_fn)


class TestMultiBenchmarkData:
    def test_def_benchmark_data_stores_fn(self):
        """def_benchmark_data(fn) stores the function."""
        from brainevent._op.main import XLACustomKernel
        from brainevent._op.benchmark import BenchmarkConfig
        prim = XLACustomKernel('_test_stores_fn')
        fn = lambda *, platform: [BenchmarkConfig("default", ())]
        prim.def_benchmark_data(fn)
        assert prim._benchmark_data_fn is fn

    def test_def_benchmark_data_none_initially(self):
        """_benchmark_data_fn is None when no fn registered."""
        from brainevent._op.main import XLACustomKernel
        prim = XLACustomKernel('_test_none_initially')
        assert prim._benchmark_data_fn is None

    def test_def_benchmark_data_overwrite(self):
        """Second def_benchmark_data call overwrites the first."""
        from brainevent._op.main import XLACustomKernel
        from brainevent._op.benchmark import BenchmarkConfig
        prim = XLACustomKernel('_test_overwrite')
        fn1 = lambda *, platform: [BenchmarkConfig("a", ())]
        fn2 = lambda *, platform: [BenchmarkConfig("b", ())]
        prim.def_benchmark_data(fn1)
        prim.def_benchmark_data(fn2)
        assert prim._benchmark_data_fn is fn2

    def test_benchmark_fn_returns_list(self):
        """Benchmark data fn should return list of BenchmarkConfig instances."""
        from brainevent._op.main import XLACustomKernel
        from brainevent._op.benchmark import BenchmarkConfig
        prim = XLACustomKernel('_test_returns_list')

        def bench_fn(*, platform):
            return [
                BenchmarkConfig("config_a", (1, 2), {"key": "val"}),
                BenchmarkConfig("config_b", (3, 4), {"key": "val2"}),
            ]

        prim.def_benchmark_data(bench_fn)
        configs = prim._benchmark_data_fn(platform='cpu')
        assert len(configs) == 2
        assert configs[0].name == "config_a"
        assert configs[1].name == "config_b"
        for config in configs:
            assert isinstance(config, BenchmarkConfig)
            assert isinstance(config.name, str)
            assert isinstance(config.args, tuple)
            assert isinstance(config.kwargs, dict)


class TestMultiBenchmarkIntegration:
    def test_real_primitive_has_benchmark_data(self):
        """Real primitives should have a benchmark data fn."""
        import brainevent
        prim = brainevent.binary_csrmv_p
        assert prim._benchmark_data_fn is not None
        assert callable(prim._benchmark_data_fn)

    def test_benchmark_fn_returns_configs(self):
        """Benchmark data fn should return list of BenchmarkConfig instances."""
        import brainevent
        from brainevent._op.benchmark import BenchmarkConfig
        prim = brainevent.binary_csrmv_p
        configs = prim._benchmark_data_fn(platform='cpu')
        assert isinstance(configs, list)
        assert len(configs) >= 1
        for config in configs:
            assert isinstance(config, BenchmarkConfig)
            assert isinstance(config.name, str)
            assert isinstance(config.args, tuple)
            assert isinstance(config.kwargs, dict)
            assert 'shape' in config.kwargs

    def test_registry_primitives_benchmark_data_structure(self):
        """All primitives with benchmark data should return proper BenchmarkConfig lists."""
        from brainevent._registry import get_registry
        from brainevent._op.benchmark import BenchmarkConfig
        registry = get_registry()
        for name, prim in registry.items():
            if prim._benchmark_data_fn is None:
                continue
            assert callable(prim._benchmark_data_fn), (
                f"Primitive '{name}': _benchmark_data_fn should be callable"
            )
            configs = prim._benchmark_data_fn(platform='cpu')
            assert isinstance(configs, list), (
                f"Primitive '{name}': benchmark data fn should return a list"
            )
            for config in configs:
                assert isinstance(config, BenchmarkConfig), (
                    f"Primitive '{name}': each entry should be a BenchmarkConfig instance"
                )
                assert isinstance(config.name, str), (
                    f"Primitive '{name}': config name should be a string"
                )
                assert isinstance(config.args, tuple), (
                    f"Primitive '{name}': args should be a tuple"
                )
                assert isinstance(config.kwargs, dict), (
                    f"Primitive '{name}': kwargs should be a dict"
                )


class TestXLACustomKernelUserDefaults:
    def test_apply_user_defaults(self, tmp_path, monkeypatch):
        """Lazy user defaults should be applied on first dispatch lookup."""
        config_path = str(tmp_path / 'brainevent' / 'defaults.json')
        monkeypatch.setattr('brainevent.config.get_config_path', lambda: config_path)
        from brainevent.config import invalidate_cache, save_user_defaults
        invalidate_cache()

        # Write a config with numba as default for cpu
        save_user_defaults({'binary_csrmv': {'cpu': 'numba'}})
        invalidate_cache()

        import brainevent
        prim = brainevent.binary_csrmv_p
        # Reset the lazy flag to test
        prim._user_defaults_applied = False
        prim._apply_user_defaults()
        assert prim._defaults.get('cpu') == 'numba'
        invalidate_cache()



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import brainstate
import jax
import jax.numpy as jnp
import numpy as np
import pytest

# Keep GPU matmul reference numerics stable (avoid TF32 drift in dense @ B checks).
if jax.default_backend() == 'gpu' and jax.config.jax_default_matmul_precision is None:
    jax.config.update('jax_default_matmul_precision', 'highest')

from brainevent._jit_scalar.float import jits, jits_p, jitsmv, jitsmv_p, jitsmm, jitsmm_p

platform = jax.default_backend()
JITS_IMPLEMENTATIONS = tuple(jits_p.available_backends(platform))
JITSMV_IMPLEMENTATIONS = tuple(jitsmv_p.available_backends(platform))
JITSMM_IMPLEMENTATIONS = tuple(jitsmm_p.available_backends(platform))


# ---- jits: transpose symmetry ----

@pytest.mark.parametrize("implementation", JITS_IMPLEMENTATIONS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_jits_transpose_symmetry(implementation, transpose, corder):
    out1 = jits(1.5, 0.1, 123, shape=(100, 50), transpose=transpose, corder=corder, backend=implementation)
    out2 = jits(1.5, 0.1, 123, shape=(100, 50), transpose=not transpose, corder=not corder, backend=implementation)
    assert jnp.allclose(out1, out2.T)
    jax.block_until_ready((out1, out2))


# ---- Forward: jitsmv (transpose=False) ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 200), (20, 100)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_forward(implementation, shape, corder):
    weight, prob, seed = 1.5, 0.1, 1234
    vector = jnp.asarray(np.random.rand(shape[1]))
    dense = jits(weight, prob, seed, shape=shape, corder=corder, backend=implementation)
    out = jitsmv(weight, prob, vector, seed=seed, shape=shape, corder=corder, backend=implementation)
    expected = dense @ vector
    assert jnp.allclose(out, expected, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, dense, out, expected))


# ---- Forward: jitsmv (transpose=True) ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 200), (20, 100)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_transpose_forward(implementation, shape, corder):
    weight, prob, seed = 1.5, 0.1, 1234
    vector = jnp.asarray(np.random.rand(shape[0]))
    dense = jits(weight, prob, seed, shape=shape, transpose=True, corder=corder, backend=implementation)
    out = jitsmv(weight, prob, vector, seed=seed, shape=shape, transpose=True, corder=corder, backend=implementation)
    expected = dense @ vector
    assert jnp.allclose(out, expected, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vector, dense, out, expected))


# ---- Forward: jitsmv zero weight ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('transpose', [True, False])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_zero_weight(implementation, transpose, corder):
    shape = (2, 3)
    v = brainstate.random.rand(shape[0]) if transpose else brainstate.random.rand(shape[1])
    result = jitsmv(0.0, 0.5, v, seed=1234, shape=shape, transpose=transpose, corder=corder, backend=implementation)
    expected = jnp.zeros(shape[1]) if transpose else jnp.zeros(shape[0])
    assert jnp.allclose(result, expected)
    jax.block_until_ready((v, result, expected))


# ---- Forward: jitsmm (transpose=False) ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 200), (20, 100)])
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmm_forward(implementation, shape, k, corder):
    weight, prob, seed = 1.5, 0.1, 1234
    B = jnp.asarray(np.random.rand(shape[1], k))
    dense = jits(weight, prob, seed, shape=shape, corder=corder, backend=implementation)
    out = jitsmm(weight, prob, B, seed=seed, shape=shape, corder=corder, backend=implementation)
    expected = dense @ B
    assert jnp.allclose(out, expected, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((B, dense, out, expected))


# ---- Forward: jitsmm (transpose=True) ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 200), (20, 100)])
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmm_transpose_forward(implementation, shape, k, corder):
    weight, prob, seed = 1.5, 0.1, 1234
    B = jnp.asarray(np.random.rand(shape[0], k))
    dense = jits(weight, prob, seed, shape=shape, transpose=True, corder=corder, backend=implementation)
    out = jitsmm(weight, prob, B, seed=seed, shape=shape, transpose=True, corder=corder, backend=implementation)
    expected = dense @ B
    assert jnp.allclose(out, expected, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((B, dense, out, expected))


# ---- Gradient JVP: jitsmv ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_jitsmv_jvp(implementation, shape, corder, transpose):
    weight, prob, seed = 1.5, 0.1, 1234
    vec_size = shape[0] if transpose else shape[1]
    x = jnp.asarray(np.random.rand(vec_size))
    dense = jits(1.0, prob, seed, shape=shape, transpose=transpose, corder=corder, backend=implementation)

    def f_fn(x, w):
        return jitsmv(w, prob, x, seed=seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    def f_dense(x, w):
        return (dense * w @ x).sum()

    w_arr = jnp.array(weight)
    t_x = jnp.ones_like(x)
    t_w = jnp.array(1.0)
    out1, jvp1 = jax.jvp(f_fn, (x, w_arr), (t_x, t_w))
    out2, jvp2 = jax.jvp(f_dense, (x, w_arr), (t_x, t_w))
    assert jnp.allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(jvp1, jvp2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((x, dense, w_arr, t_x, t_w, out1, jvp1, out2, jvp2))


# ---- Gradient VJP: jitsmv ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_jitsmv_vjp(implementation, shape, corder, transpose):
    weight, prob, seed = 1.5, 0.1, 1234
    vec_size = shape[0] if transpose else shape[1]
    x = jnp.asarray(np.random.rand(vec_size))
    dense = jits(1.0, prob, seed, shape=shape, transpose=transpose, corder=corder, backend=implementation)

    def f_fn(x, w):
        return jitsmv(w, prob, x, seed=seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    def f_dense(x, w):
        return (dense * w @ x).sum()

    w_arr = jnp.array(weight)
    out1, (vjp_x1, vjp_w1) = jax.value_and_grad(f_fn, argnums=(0, 1))(x, w_arr)
    out2, (vjp_x2, vjp_w2) = jax.value_and_grad(f_dense, argnums=(0, 1))(x, w_arr)
    assert jnp.allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(vjp_x1, vjp_x2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(vjp_w1, vjp_w2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((x, dense, w_arr, out1, vjp_x1, vjp_w1, out2, vjp_x2, vjp_w2))


# ---- Gradient JVP: jitsmm ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_jitsmm_jvp(implementation, k, shape, corder, transpose):
    weight, prob, seed = 1.5, 0.1, 1234
    mat_rows = shape[0] if transpose else shape[1]
    X = jnp.asarray(np.random.rand(mat_rows, k))
    dense = jits(1.0, prob, seed, shape=shape, transpose=transpose, corder=corder, backend=implementation)

    def f_fn(X, w):
        return jitsmm(w, prob, X, seed=seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    def f_dense(X, w):
        return (dense * w @ X).sum()

    w_arr = jnp.array(weight)
    t_X = jnp.ones_like(X)
    t_w = jnp.array(1.0)
    out1, jvp1 = jax.jvp(f_fn, (X, w_arr), (t_X, t_w))
    out2, jvp2 = jax.jvp(f_dense, (X, w_arr), (t_X, t_w))
    assert jnp.allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(jvp1, jvp2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((X, dense, w_arr, t_X, t_w, out1, jvp1, out2, jvp2))


# ---- Gradient VJP: jitsmm ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('k', [10])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
@pytest.mark.parametrize('transpose', [True, False])
def test_jitsmm_vjp(implementation, k, shape, corder, transpose):
    weight, prob, seed = 1.5, 0.1, 1234
    mat_rows = shape[0] if transpose else shape[1]
    X = jnp.asarray(np.random.rand(mat_rows, k))
    dense = jits(1.0, prob, seed, shape=shape, transpose=transpose, corder=corder, backend=implementation)

    def f_fn(X, w):
        return jitsmm(w, prob, X, seed=seed, shape=shape, transpose=transpose, corder=corder,
                      backend=implementation).sum()

    def f_dense(X, w):
        return (dense * w @ X).sum()

    w_arr = jnp.array(weight)
    out1, (vjp_x1, vjp_w1) = jax.value_and_grad(f_fn, argnums=(0, 1))(X, w_arr)
    out2, (vjp_x2, vjp_w2) = jax.value_and_grad(f_dense, argnums=(0, 1))(X, w_arr)
    assert jnp.allclose(out1, out2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(vjp_x1, vjp_x2, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(vjp_w1, vjp_w2, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((X, dense, w_arr, out1, vjp_x1, vjp_w1, out2, vjp_x2, vjp_w2))


# ---- Batching: jitsmv over vectors ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_vmap_over_vectors(implementation, batch_size, shape, corder):
    weight, prob, seed = 1.05, 0.1, 123
    vectors = brainstate.random.rand(batch_size, shape[1])

    def f(vector):
        return jitsmv(weight, prob, vector, seed=seed, shape=shape, corder=corder, backend=implementation)

    results = jax.vmap(f)(vectors)
    assert results.shape == (batch_size, shape[0])

    results_loop = brainstate.transform.for_loop(f, vectors)
    assert results_loop.shape == (batch_size, shape[0])

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vectors, results, results_loop))


# ---- Batching: jitsmv over vectors (transpose) ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_transpose_vmap_over_vectors(implementation, batch_size, shape, corder):
    weight, prob, seed = 1.05, 0.1, 123
    vectors = brainstate.random.rand(batch_size, shape[0])

    def f(vector):
        return jitsmv(weight, prob, vector, seed=seed, shape=shape, transpose=True, corder=corder,
                      backend=implementation)

    results = jax.vmap(f)(vectors)
    assert results.shape == (batch_size, shape[1])

    results_loop = brainstate.transform.for_loop(f, vectors)
    assert results_loop.shape == (batch_size, shape[1])

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((vectors, results, results_loop))


# ---- Batching: jitsmv over weight ----

@pytest.mark.parametrize("implementation", JITSMV_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmv_vmap_over_weight(implementation, batch_size, shape, corder):
    prob, seed = 0.1, 123
    weights = brainstate.random.rand(batch_size)
    vector = brainstate.random.rand(shape[1])

    def f(w):
        return jitsmv(w, prob, vector, seed=seed, shape=shape, corder=corder, backend=implementation)

    results = jax.vmap(f)(weights)
    assert results.shape == (batch_size, shape[0])

    results_loop = brainstate.transform.for_loop(f, weights)
    assert results_loop.shape == (batch_size, shape[0])

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((weights, vector, results, results_loop))


# ---- Batching: jitsmm over matrices ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('k', [5])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmm_vmap_over_matrices(implementation, batch_size, k, shape, corder):
    weight, prob, seed = 1.05, 0.1, 123
    matrices = brainstate.random.rand(batch_size, shape[1], k)

    def f(mat):
        return jitsmm(weight, prob, mat, seed=seed, shape=shape, corder=corder, backend=implementation)

    outs = jax.vmap(f)(matrices)
    assert outs.shape == (batch_size, shape[0], k)

    outs_loop = brainstate.transform.for_loop(f, matrices)
    assert outs_loop.shape == (batch_size, shape[0], k)

    assert jnp.allclose(outs, outs_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((matrices, outs, outs_loop))


# ---- Batching: jitsmm over matrices (transpose) ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('k', [5])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmm_transpose_vmap_over_matrices(implementation, batch_size, k, shape, corder):
    weight, prob, seed = 1.05, 0.1, 123
    matrices = brainstate.random.rand(batch_size, shape[0], k)

    def f(mat):
        return jitsmm(weight, prob, mat, seed=seed, shape=shape, transpose=True, corder=corder, backend=implementation)

    outs = jax.vmap(f)(matrices)
    assert outs.shape == (batch_size, shape[1], k)

    outs_loop = brainstate.transform.for_loop(f, matrices)
    assert outs_loop.shape == (batch_size, shape[1], k)

    assert jnp.allclose(outs, outs_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((matrices, outs, outs_loop))


# ---- Batching: jitsmm over weight ----

@pytest.mark.parametrize("implementation", JITSMM_IMPLEMENTATIONS)
@pytest.mark.parametrize('batch_size', [10])
@pytest.mark.parametrize('k', [5])
@pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
@pytest.mark.parametrize('corder', [True, False])
def test_jitsmm_vmap_over_weight(implementation, batch_size, k, shape, corder):
    prob, seed = 0.1, 123
    weights = brainstate.random.rand(batch_size)
    matrix = brainstate.random.rand(shape[1], k)

    def f(w):
        return jitsmm(w, prob, matrix, seed=seed, shape=shape, corder=corder, backend=implementation)

    results = jax.vmap(f)(weights)
    assert results.shape == (batch_size, shape[0], k)

    results_loop = brainstate.transform.for_loop(f, weights)
    assert results_loop.shape == (batch_size, shape[0], k)

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((weights, matrix, results, results_loop))


# ---- Batching: jits over weight ----

@pytest.mark.parametrize("implementation", JITS_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 50)])
def test_jits_vmap_over_weight(implementation, shape):
    prob, seed = 0.1, 123

    def f(weight):
        return jits(weight, prob, seed, shape=shape, backend=implementation)

    weights = brainstate.random.rand(10)
    results = jax.vmap(f)(weights)
    assert results.shape == (10,) + shape

    results_loop = brainstate.transform.for_loop(f, weights)
    assert results_loop.shape == (10,) + shape

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((weights, results, results_loop))


# ---- Batching: jits over prob ----

@pytest.mark.parametrize("implementation", JITS_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 50)])
def test_jits_vmap_over_prob(implementation, shape):
    weight, seed = 1.5, 123

    def f(prob):
        return jits(weight, prob, seed, shape=shape, backend=implementation)

    probs = brainstate.random.rand(10)
    results = jax.vmap(f)(probs)
    assert results.shape == (10,) + shape

    results_loop = brainstate.transform.for_loop(f, probs)
    assert results_loop.shape == (10,) + shape

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((probs, results, results_loop))


# ---- Batching: jits over seed ----

@pytest.mark.parametrize("implementation", JITS_IMPLEMENTATIONS)
@pytest.mark.parametrize('shape', [(100, 50)])
def test_jits_vmap_over_seed(implementation, shape):
    weight, prob = 1.5, 0.1

    def f(seed):
        return jits(weight, prob, seed, shape=shape, backend=implementation)

    seeds = brainstate.random.randint(0, 100000, 10)
    results = jax.vmap(f)(seeds)
    assert results.shape == (10,) + shape

    results_loop = brainstate.transform.for_loop(f, seeds)
    assert results_loop.shape == (10,) + shape

    assert jnp.allclose(results, results_loop, rtol=1e-4, atol=1e-4)
    jax.block_until_ready((seeds, results, results_loop))
